// Copyright © 2019-2020 Nicolas Mailhot <nim@fedoraproject.org>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package main

import (
  "errors"
  "fmt"
  "os"
  "pagure.io/modist/filter"
  "pagure.io/modist/index"
  "pagure.io/modist/install"
  "pagure.io/modist/list"
  "pagure.io/modist/pack"
  "pagure.io/modist/version"
  "github.com/jawher/mow.cli"
)

func check(err error) {
  if err != nil {
    fmt.Fprintf(os.Stderr,"Error: %v\n", err)
    cli.Exit(1)
  }
}

var ErrUnknownOutputFormat = errors.New("unknown output format")

func main() {
  app := cli.App("modist", "A small utility to create and query Go (Golang) modules files.")

  app.Version("V version", "0.0.1")

  app.Spec = ""

  // Arguments and options that may be used by several subcommands
  modVerSetByUser := false
  modVerProto := cli.StringOpt{
  Name:       "v version",
  Desc:       "use VERSION as semantic version core string",
    SetByUser:  &modVerSetByUser,
  }

  modPreSetByUser := false
  modPreProto := cli.StringOpt{
  Name:       "P pre-release",
  Desc:       `use PRE as semantic version pre-release string
               (PRE should not include VERSION in any form)`,
    SetByUser:  &modPreSetByUser,
  }

  modCommitSetByUser := false
  modCommitProto := cli.StringOpt{
  Name:       "C commit",
  Desc:       "use COMMIT identifier (usually, a hash).",
    SetByUser:  &modCommitSetByUser,
  }

  modTimeSetByUser := false
  modTimeProto := cli.StringOpt{
  Name:       "H time",
  Desc:       `use source TIMESTAMP, in RFC 3339 format
               (as produced by “date -u --iso-8601=seconds”)`,
  SetByUser:  &modTimeSetByUser,
  }

  modBuildSetByUser := false
  modBuildProto := cli.StringOpt{
  Name:       "B build",
  Desc:       `use BUILD as semantic version build string
               (release in RPM conventions)`,
  SetByUser:  &modBuildSetByUser,
  }

  modEpochSetByUser := false
  modEpochProto := cli.IntOpt{
  Name:       "Z epoch",
  Desc:       `use versioning EPOCH`,
  SetByUser:  &modEpochSetByUser,
  }

  includeModsSetByUser := false
  includeModsProto := cli.StringsOpt{
  Name:       "M include-module",
  Desc:       `process this module
               if not set, will process all found modules`,
    SetByUser: &includeModsSetByUser,
  }

  ignoreModsSetByUser := false
  ignoreModsProto := cli.StringsOpt{
  Name:       "m ignore-module",
  Desc:       "ignore this module",
    SetByUser: &ignoreModsSetByUser,
  }

  includeFilesSetByUser := false
  includeFilesProto := cli.StringsOpt{
  Name:       "F include-file",
  Desc:       "include files with an exact path match",
    SetByUser: &includeFilesSetByUser,
  }

  includeExtensionsSetByUser := false
  includeExtensionsProto := cli.StringsOpt{
  Name:       "E include-extension",
  Desc:       "include files ending in this extension",
    SetByUser: &includeExtensionsSetByUser,
  }

  includeDirectoriesSetByUser := false
  includeDirectoriesProto := cli.StringsOpt{
  Name:       "D include-directory",
  Desc:       `include files contained in the directory itself
               do not include files contained in its subdirectories`,
    SetByUser: &includeDirectoriesSetByUser,
  }

  includeTreesSetByUser := false
  includeTreesProto := cli.StringsOpt{
  Name:       "T include-tree",
  Desc:       `include files contained in the tree root and all its
               subdirectories`,
    SetByUser: &includeTreesSetByUser,
  }

  includeRegexpsSetByUser := false
  includeRegexpsProto := cli.StringsOpt{
  Name:       "R include-regexp",
  Desc:       "include files matching this regexp",
    SetByUser: &includeRegexpsSetByUser,
  }

  ignoreFilesSetByUser := false
  ignoreFilesProto := cli.StringsOpt{
  Name:       "f ignore-file",
  Desc:       "ignore files with an exact path match",
    SetByUser: &ignoreFilesSetByUser,
  }

  ignoreExtensionsSetByUser := false
  ignoreExtensionsProto := cli.StringsOpt{
  Name:       "e ignore-extension",
  Desc:       "ignore files ending in this extension",
    SetByUser: &ignoreExtensionsSetByUser,
  }

  ignoreDirectoriesSetByUser := false
  ignoreDirectoriesProto := cli.StringsOpt{
  Name:       "d ignore-directory",
  Desc:       `ignore files contained in the directory itself
               do not ignore files contained in its subdirectories`,
    SetByUser: &ignoreDirectoriesSetByUser,
  }

  ignoreTreesSetByUser := false
  ignoreTreesProto := cli.StringsOpt{
  Name:       "t ignore-tree",
  Desc:       `ignore files contained in the tree root and all its
               subdirectories`,
    SetByUser: &ignoreTreesSetByUser,
  }

  ignoreRegexpsSetByUser := false
  ignoreRegexpsProto := cli.StringsOpt{
  Name:       "r ignore-regexp",
  Desc:       "ignore files matching this regexp",
    SetByUser: &ignoreRegexpsSetByUser,
  }

  quietSetByUser := false
  quietProto := cli.BoolOpt{
  Name:       "q quiet",
  Value:      false,
  Desc:       "limit messages to the minimum",
    SetByUser: &quietSetByUser,
  }

  outputSetByUser := false
  outputProto := cli.StringOpt{
  Name:       "o output",
  Desc:       `report the result of the command to output file
               the report is not affected by DESTDIR or PREFIX`,
    SetByUser: &outputSetByUser,
  }

  outputFormatSetByUser := false
  outputFormatProto := cli.StringOpt{
  Name:       "O format",
  Value:      "rpm",
  Desc:       "use the selected output format",
    SetByUser: &outputFormatSetByUser,
  }

  prefixSetByUser := false
  prefixProto := cli.StringOpt{
  Name:       "p prefix",
  Desc:       `pretend writing into DESTDIR
               actually write into PREFIX/DESTDIR`,
    SetByUser: &prefixSetByUser,
  }

  sourcesSetByUser := false
  sourcesProto := cli.StringsArg{
    Name:      "SOURCE",
    Desc:      `a Go module descriptor file, OR
                a directory that will be scanned for Go modules`,
    SetByUser: &sourcesSetByUser,
  }

  dstDirSetByUser := false
  dstDirProto := cli.StringArg{
  Name:       "DESTDIR",
  Desc:       "write files into DESTDIR",
    SetByUser: &dstDirSetByUser,
  }

  filterHelp := `

⚠ Filtering module content

The command will try to select sensible module content by default.

--include… and --ignore… filtering options are provided to fine-tune the default
selections. --include… options are processed first, followed by --ignore…
options (this may change in the future).

Filtering options apply to relative file paths within a module tree.`

  versionHelp := `

⚠ Go module versioning

Go module versioning has no concept of post-version identifiers:
 – PRE will identify a state anterior to VERSION,
 — COMMIT alone will identify a state anterior to any pre-version of VERSION,
 — COMMIT and PRE will identify a state, anterior to VERSION, preceded by a
   series of pre-versions, ending in PRE.

When processing COMMITs and PREs, take care to increment VERSION so it
identifies the first version that will be released after those states.

⚠ RPM versioning

Since RPM versions can end up in filenames on case-insensitive filesystems or
file servers, modist will refuse to honor PRE ordering that relies on
casing.`

  app.Command("provides", "Print Go module provides", func(cmd *cli.Cmd) {

    onlyNameSetByUser := false
    onlyNameProto := cli.BoolOpt{
    Name:       "n only-name",
    Value:      false,
    Desc:       "limit report to the module name",
      SetByUser: &onlyNameSetByUser,
    }

    onlyVersionSetByUser := false
    onlyVersionProto := cli.BoolOpt{
    Name:       "only-version",
    Value:      false,
    Desc:       "limit report to the module version",
      SetByUser: &onlyVersionSetByUser,
    }

    cmd.Spec =  "[-v|--version=<VERSION>] " +
                "[-P|--pre-release=<PRE>] " +
               "[[-C|--commit=<COMMIT>] (-H|--time=<TIMESTAMP>)] " +
                "[-B|--build=<BUILD>] " +
                "[-Z|--epoch=<EPOCH>] " +
               "[[-M|--include-module=<module name>] | " +
                "[-m|--ignore-module=<module name>]]... " +
                "[-n|--only-name|--only-version] " +
                "[-o|--output=<file path>] [-O|--format=<rpm|go>] " +
                "[-q|--quiet] SOURCE..."

    cmd.LongDesc = "Print what the Go modules located in SOURCE… provide." + versionHelp

    modVer       := cmd.String(modVerProto)
    modPre       := cmd.String(modPreProto)
    modCommit    := cmd.String(modCommitProto)
    modTime      := cmd.String(modTimeProto)
    modBuild     := cmd.String(modBuildProto)
    modEpoch     := cmd.Int(modEpochProto)
    includeMods  := cmd.Strings(includeModsProto)
    ignoreMods   := cmd.Strings(ignoreModsProto)
    onlyName     := cmd.Bool(onlyNameProto)
    onlyVersion  := cmd.Bool(onlyVersionProto)
    output       := cmd.String(outputProto)
    outputFormat := cmd.String(outputFormatProto)
    quiet        := cmd.Bool(quietProto)
    sources      := cmd.Strings(sourcesProto)

    cmd.Action = func() {
      var err error
      var ver *version.Version
      if modVerSetByUser || modPreSetByUser || modCommitSetByUser {
        ver, err = version.New(*modVer, *modPre, *modCommit, *modTime,
                               *modBuild, *modEpoch, *quiet)
        check(err)
      }
      _, _, err  = list.Provides(*sources, ver, *includeMods, *ignoreMods,
                                 *onlyName, *onlyVersion,
                                 *output, *outputFormat, *quiet)
      check(err)
    }
  })

  app.Command("requires", "Print Go module requires", func(cmd *cli.Cmd) {

    cmd.Spec = "[[-M|--include-module=<module name>] | " +
                "[-m|--ignore-module=<module name>]]... " +
                "[-o|--output=<file path>] [-O|--format=<rpm|go>] " +
                "[-q|--quiet] SOURCE..."

    cmd.LongDesc = "Print what the Go modules located in SOURCE… require."

    includeMods  := cmd.Strings(includeModsProto)
    ignoreMods   := cmd.Strings(ignoreModsProto)
    output       := cmd.String(outputProto)
    outputFormat := cmd.String(outputFormatProto)
    quiet        := cmd.Bool(quietProto)
    sources      := cmd.Strings(sourcesProto)

    cmd.Action = func() {
      _, _, err := list.Requires(*sources, *includeMods, *ignoreMods,
                                 *output, *outputFormat, *quiet)
      check(err)
    }
  })

  app.Command("buildrequires", "Print Go module build requires", func(cmd *cli.Cmd) {

    cmd.Spec =  "[-v|--version=<VERSION>] " +
                "[-P|--pre-release=<PRE>] " +
               "[[-C|--commit=<COMMIT>] (-H|--time=<TIMESTAMP>)] " +
                "[-B|--build=<BUILD>] " +
                "[-Z|--epoch=<EPOCH>] " +
               "[[-M|--include-module=<module name>] | " +
                "[-m|--ignore-module=<module name>]]... " +
                "[-o|--output=<file path>] [-O|--format=<rpm|go>] " +
                "[-q|--quiet] SOURCE..."

    brHelp := `

⚠ Go module requires and build requires differences

A Go module can declare replacements and exclusions that only apply to the
module own build. Therefore a module build requirements, may differ from the
requirements it exports to other modules.

This difference may cause problems, if other modules attempt to build the code
of the original module, with incompatible requirements, that the module masked
in its own build.

modist computes the build requirements of a synchronized set of modules with
the same versionning. Attempting to build non-synchronized sets of
modules in a single pass is not supported.`

    cmd.LongDesc = "Print what building the Go modules located in SOURCE… requires." + brHelp + versionHelp

    modVer       := cmd.String(modVerProto)
    modPre       := cmd.String(modPreProto)
    modCommit    := cmd.String(modCommitProto)
    modTime      := cmd.String(modTimeProto)
    modBuild     := cmd.String(modBuildProto)
    modEpoch     := cmd.Int(modEpochProto)
    includeMods  := cmd.Strings(includeModsProto)
    ignoreMods   := cmd.Strings(ignoreModsProto)
    output       := cmd.String(outputProto)
    outputFormat := cmd.String(outputFormatProto)
    quiet        := cmd.Bool(quietProto)
    sources      := cmd.Strings(sourcesProto)

    cmd.Action = func() {
      var err error
      var ver *version.Version
      if modVerSetByUser || modPreSetByUser || modCommitSetByUser {
        ver, err = version.New(*modVer, *modPre, *modCommit, *modTime,
                               *modBuild, *modEpoch, *quiet)
        check(err)
      }
      _, _, err  = list.BuildRequires(*sources, ver, *includeMods, *ignoreMods,
                                      *output, *outputFormat, *quiet)
      check(err)
    }
  })

  app.Command("pathprovides", "Print Go module GOPATH provides", func(cmd *cli.Cmd) {

    onlyNameSetByUser := false
    onlyNameProto := cli.BoolOpt{
    Name:       "n only-name",
    Value:      false,
    Desc:       "limit report to the module name",
      SetByUser: &onlyNameSetByUser,
    }

    onlyVersionSetByUser := false
    onlyVersionProto := cli.BoolOpt{
    Name:       "only-version",
    Value:      false,
    Desc:       "limit report to the module version",
      SetByUser: &onlyVersionSetByUser,
    }

    cmd.Spec =  "[-v|--version=<VERSION>] " +
                "[-P|--pre-release=<PRE>] " +
               "[[-C|--commit=<COMMIT>] (-H|--time=<TIMESTAMP>)] " +
                "[-B|--build=<BUILD>] " +
                "[-Z|--epoch=<EPOCH>] " +
               "[[-M|--include-module=<module name>] | " +
                "[-m|--ignore-module=<module name>]]... " +
               "[[-F|--include-file=<in-module file path>] | " +
                "[-E|--include-extension=<value>] | " +
                "[-D|--include-directory=<in-module directory path>] | " +
                "[-T|--include-tree=<in-module tree root path>] | " +
                "[-R|--include-regexp=<in-module path regexp>] | " +
                "[-f|--ignore-file=<in-module file path>] | " +
                "[-e|--ignore-extension=<value>] | " +
                "[-d|--ignore-directory=<in-module directory path>] | " +
                "[-t|--ignore-tree=<in-module tree root path>] | " +
                "[-r|--ignore-regexp=<in-module path regexp>]]... " +
                "[-n|--only-name|--only-version] " +
                "[-o|--output=<file path>] [-O|--format=<rpm|go>] " +
                "[-q|--quiet] SOURCE..."

    cmd.LongDesc = "Print what the Go modules located in SOURCE… provide in GOPATH mode." + versionHelp

    modVer             := cmd.String(modVerProto)
    modPre             := cmd.String(modPreProto)
    modCommit          := cmd.String(modCommitProto)
    modTime            := cmd.String(modTimeProto)
    modBuild           := cmd.String(modBuildProto)
    modEpoch           := cmd.Int(modEpochProto)
    includeMods        := cmd.Strings(includeModsProto)
    ignoreMods         := cmd.Strings(ignoreModsProto)
    includeFiles       := cmd.Strings(includeFilesProto)
    includeExtensions  := cmd.Strings(includeExtensionsProto)
    includeDirectories := cmd.Strings(includeDirectoriesProto)
    includeTrees       := cmd.Strings(includeTreesProto)
    includeRegexps     := cmd.Strings(includeRegexpsProto)
    ignoreFiles        := cmd.Strings(ignoreFilesProto)
    ignoreExtensions   := cmd.Strings(ignoreExtensionsProto)
    ignoreDirectories  := cmd.Strings(ignoreDirectoriesProto)
    ignoreTrees        := cmd.Strings(ignoreTreesProto)
    ignoreRegexps      := cmd.Strings(ignoreRegexpsProto)
    onlyName           := cmd.Bool(onlyNameProto)
    onlyVersion        := cmd.Bool(onlyVersionProto)
    output             := cmd.String(outputProto)
    outputFormat       := cmd.String(outputFormatProto)
    quiet              := cmd.Bool(quietProto)
    sources            := cmd.Strings(sourcesProto)

    cmd.Action = func() {
      ver, err := version.New(*modVer, *modPre, *modCommit, *modTime,
                              *modBuild, *modEpoch, *quiet)
      check(err)
      filters      := filter.Default
      filters, err  = filters.AddByType(*includeFiles, *includeExtensions,
                                        *includeDirectories, *includeTrees,
                                        *includeRegexps,
                                        *ignoreFiles,  *ignoreExtensions,
                                        *ignoreDirectories,  *ignoreTrees,
                                        *ignoreRegexps)
      check(err)
      _, _, err  = filter.PathProvides(*sources, ver, *includeMods, *ignoreMods,
                                       filters, *onlyName, *onlyVersion,
                                       *output, *outputFormat, *quiet)
      check(err)
    }
  })

  app.Command("pathrequires", "Print Go module GOPATH requires", func(cmd *cli.Cmd) {

    cmd.Spec = "[[-M|--include-module=<module name>] | " +
                "[-m|--ignore-module=<module name>]]... " +
                "[-o|--output=<file path>] [-O|--format=<rpm|go>] " +
                "[-q|--quiet] SOURCE..."

    cmd.LongDesc = "Print what the Go modules located in SOURCE… require in GOPATH mode."

    includeMods  := cmd.Strings(includeModsProto)
    ignoreMods   := cmd.Strings(ignoreModsProto)
    output       := cmd.String(outputProto)
    outputFormat := cmd.String(outputFormatProto)
    quiet        := cmd.Bool(quietProto)
    sources      := cmd.Strings(sourcesProto)

    cmd.Action = func() {
      _, _, err := list.PathRequires(*sources, *includeMods, *ignoreMods,
                                     *output, *outputFormat, *quiet)
      check(err)
    }
  })

  app.Command("version", "Compute and convert version strings", func(cmd *cli.Cmd) {

    cmd.Spec =  "[-v|--version=<VERSION>] " +
                "[-P|--pre-release=<PRE>] " +
               "[[-C|--commit=<COMMIT>] (-H|--time=<TIMESTAMP>)] " +
                "[-B|--build=<BUILD>] " +
                "[-Z|--epoch=<EPOCH>] " +
                "[-O|--format=<rpm|go>]" +
                "[-q|--quiet]"

    cmd.LongDesc = `
Compute version strings from individual components, or convert between version
formats.` + versionHelp

    modVer       := cmd.String(modVerProto)
    modPre       := cmd.String(modPreProto)
    modCommit    := cmd.String(modCommitProto)
    modTime      := cmd.String(modTimeProto)
    modBuild     := cmd.String(modBuildProto)
    modEpoch     := cmd.Int(modEpochProto)
    outputFormat := cmd.String(outputFormatProto)
    quiet        := cmd.Bool(quietProto)

    cmd.Action = func() {
      ver, err := version.New(*modVer, *modPre, *modCommit, *modTime,
                              *modBuild, *modEpoch, *quiet)
      check(err)
      result, err := ver.Format(*outputFormat)
      check(err)
      fmt.Printf("%s\n", result)
    }
  })

  app.Command("discover", "Discover unpacked Go modules", func(cmd *cli.Cmd) {

    onlyRootSetByUser := false
    onlyRootProto := cli.BoolOpt{
    Name:       "r only-root",
    Value:      false,
    Desc:       "limit report to the module file path",
      SetByUser: &onlyRootSetByUser,
    }

    outputFormatSetByUser := false
    outputFormatProto := cli.StringOpt{
    Name:       "O format",
    Value:      "txt",
    Desc:       "use the selected output format",
      SetByUser: &outputFormatSetByUser,
    }

    cmd.Spec = "[[-M|--include-module=<module name>] | " +
                "[-m|--ignore-module=<module name>]]... " +
                "[-r|--only-root] " +
                "[-o|--output=<file path>] [-O|--format=<txt|csv>] " +
                "[-q|--quiet] SOURCE..."

    cmd.LongDesc = "Discover the Go modules located in SOURCE…"

    includeMods  := cmd.Strings(includeModsProto)
    ignoreMods   := cmd.Strings(ignoreModsProto)
    onlyRoot     := cmd.Bool(onlyRootProto)
    output       := cmd.String(outputProto)
    outputFormat := cmd.String(outputFormatProto)
    quiet        := cmd.Bool(quietProto)
    sources      := cmd.Strings(sourcesProto)

    cmd.Action = func() {
      _, _, err := list.Discover(*sources, *includeMods, *ignoreMods,
                                 *onlyRoot, *output, *outputFormat, *quiet)
      check(err)
    }
  })

  app.Command("list", "List unpacked Go modules", func(cmd *cli.Cmd) {

    outputFormatSetByUser := false
    outputFormatProto := cli.StringOpt{
    Name:       "O format",
    Value:      "txt",
    Desc:       "use the selected output format",
      SetByUser: &outputFormatSetByUser,
    }

    cmd.LongDesc = "List the unpacked Go modules located in SOURCE…" + filterHelp

    cmd.Spec = "[[-M|--include-module=<module name>] | " +
                "[-m|--ignore-module=<module name>]]... " +
               "[[-F|--include-file=<in-module file path>] | " +
                "[-E|--include-extension=<value>] | " +
                "[-D|--include-directory=<in-module directory path>] | " +
                "[-T|--include-tree=<in-module tree root path>] | " +
                "[-R|--include-regexp=<in-module path regexp>] | " +
                "[-f|--ignore-file=<in-module file path>] | " +
                "[-e|--ignore-extension=<value>] | " +
                "[-d|--ignore-directory=<in-module directory path>] | " +
                "[-t|--ignore-tree=<in-module tree root path>] | " +
                "[-r|--ignore-regexp=<in-module path regexp>]]... " +
                "[-o|--output=<file path>] [-O|--format=<txt|csv>] " +
                "[-q|--quiet] SOURCE..."

    includeMods        := cmd.Strings(includeModsProto)
    ignoreMods         := cmd.Strings(ignoreModsProto)
    includeFiles       := cmd.Strings(includeFilesProto)
    includeExtensions  := cmd.Strings(includeExtensionsProto)
    includeDirectories := cmd.Strings(includeDirectoriesProto)
    includeTrees       := cmd.Strings(includeTreesProto)
    includeRegexps     := cmd.Strings(includeRegexpsProto)
    ignoreFiles        := cmd.Strings(ignoreFilesProto)
    ignoreExtensions   := cmd.Strings(ignoreExtensionsProto)
    ignoreDirectories  := cmd.Strings(ignoreDirectoriesProto)
    ignoreTrees        := cmd.Strings(ignoreTreesProto)
    ignoreRegexps      := cmd.Strings(ignoreRegexpsProto)
    output             := cmd.String(outputProto)
    outputFormat       := cmd.String(outputFormatProto)
    quiet              := cmd.Bool(quietProto)
    sources            := cmd.Strings(sourcesProto)

    cmd.Action = func() {
      filters      := filter.Default
      filters, err := filters.AddByType(*includeFiles, *includeExtensions,
                                        *includeDirectories, *includeTrees,
                                        *includeRegexps,
                                        *ignoreFiles,  *ignoreExtensions,
                                        *ignoreDirectories,  *ignoreTrees,
                                        *ignoreRegexps)
      check(err)
      _, _, err = filter.List(*sources, nil, *includeMods, *ignoreMods, filters,
                              *output, *outputFormat, true, *quiet)
      check(err)
    }
  })

  app.Command("install", "Install unpacked Go modules", func(cmd *cli.Cmd) {

    qualifyRootSetByUser := false
    qualifyRootProto := cli.BoolOpt{
    Name:       "Q qualify-root",
    Value:      false,
    Desc:       `qualify the module root with the module version
                 (~/go/pkg/mod behaviour)`,
      SetByUser: &qualifyRootSetByUser,
    }

    outputFormatSetByUser := false
    outputFormatProto := cli.StringOpt{
    Name:       "O format",
    Value:      "txt",
    Desc:       "use the selected output format",
      SetByUser: &outputFormatSetByUser,
    }

    cmd.LongDesc = "Install the unpacked Go modules located in SOURCE… into DESTDIR." + filterHelp + versionHelp

    cmd.Spec =  "[-v|--version=<VERSION>] " +
                "[-P|--pre-release=<PRE>] " +
                "[[-C|--commit=<COMMIT>] (-H|--time=<TIMESTAMP>)] " +
                "[-B|--build=<BUILD>] " +
                "[-Z|--epoch=<EPOCH>] " +
                "[[-M|--include-module=<module name>] | " +
                "[-m|--ignore-module=<module name>]]... " +
               "[[-F|--include-file=<in-module file path>] | " +
                "[-E|--include-extension=<value>] | " +
                "[-D|--include-directory=<in-module directory path>] | " +
                "[-T|--include-tree=<in-module tree root path>] | " +
                "[-R|--include-regexp=<in-module path regexp>] | " +
                "[-f|--ignore-file=<in-module file path>] | " +
                "[-e|--ignore-extension=<value>] | " +
                "[-d|--ignore-directory=<in-module directory path>] | " +
                "[-t|--ignore-tree=<in-module tree root path>] | " +
                "[-r|--ignore-regexp=<in-module path regexp>]]... " +
                "[-p|--prefix=<PREFIX>] " +
                "[-Q|--qualify-root] " +
                "[-o|--output=<file path>] [-O|--format=<rpm|txt|csv>] " +
                "[-q|--quiet] SOURCE... DESTDIR"

    modVer             := cmd.String(modVerProto)
    modPre             := cmd.String(modPreProto)
    modCommit          := cmd.String(modCommitProto)
    modTime            := cmd.String(modTimeProto)
    modBuild           := cmd.String(modBuildProto)
    modEpoch           := cmd.Int(modEpochProto)
    includeMods        := cmd.Strings(includeModsProto)
    ignoreMods         := cmd.Strings(ignoreModsProto)
    includeFiles       := cmd.Strings(includeFilesProto)
    includeExtensions  := cmd.Strings(includeExtensionsProto)
    includeDirectories := cmd.Strings(includeDirectoriesProto)
    includeTrees       := cmd.Strings(includeTreesProto)
    includeRegexps     := cmd.Strings(includeRegexpsProto)
    ignoreFiles        := cmd.Strings(ignoreFilesProto)
    ignoreExtensions   := cmd.Strings(ignoreExtensionsProto)
    ignoreDirectories  := cmd.Strings(ignoreDirectoriesProto)
    ignoreTrees        := cmd.Strings(ignoreTreesProto)
    ignoreRegexps      := cmd.Strings(ignoreRegexpsProto)
    prefix             := cmd.String(prefixProto)
    qualifyRoot        := cmd.Bool(qualifyRootProto)
    output             := cmd.String(outputProto)
    outputFormat       := cmd.String(outputFormatProto)
    quiet              := cmd.Bool(quietProto)
    sources            := cmd.Strings(sourcesProto)
    dstDir             := cmd.String(dstDirProto)


    cmd.Action = func() {
      ver, err := version.New(*modVer, *modPre, *modCommit, *modTime,
                              *modBuild, *modEpoch, *quiet)
      check(err)
      filters      := filter.Default
      filters, err  = filters.AddByType(*includeFiles, *includeExtensions,
                                        *includeDirectories, *includeTrees,
                                        *includeRegexps,
                                        *ignoreFiles,  *ignoreExtensions,
                                        *ignoreDirectories,  *ignoreTrees,
                                        *ignoreRegexps)
      check(err)
      _, err = install.Scan(*sources, ver, *includeMods, *ignoreMods, filters,
                            *prefix, *qualifyRoot, *dstDir,
                            *output, *outputFormat, *quiet)
      check(err)
    }
  })

  app.Command("pack", "Pack unpacked Go modules", func(cmd *cli.Cmd) {

    outputFormatSetByUser := false
    outputFormatProto := cli.StringOpt{
    Name:       "O format",
    Value:      "txt",
    Desc:       "use the selected output format",
      SetByUser: &outputFormatSetByUser,
    }

    cmd.LongDesc = "Pack the unpacked Go modules located in SOURCE… into DESTDIR." + filterHelp + versionHelp

    cmd.Spec =  "[-v|--version=<VERSION>] " +
                "[-P|--pre-release=<PRE>] " +
               "[[-C|--commit=<COMMIT>] (-H|--time=<TIMESTAMP>)] " +
                "[-B|--build=<BUILD>] " +
                "[-Z|--epoch=<EPOCH>] " +
               "[[-M|--include-module=<module name>] | " +
                "[-m|--ignore-module=<module name>]]... " +
               "[[-F|--include-file=<in-module file path>] | " +
                "[-E|--include-extension=<value>] | " +
                "[-D|--include-directory=<in-module directory path>] | " +
                "[-T|--include-tree=<in-module tree root path>] | " +
                "[-R|--include-regexp=<in-module path regexp>] | " +
                "[-f|--ignore-file=<in-module file path>] | " +
                "[-e|--ignore-extension=<value>] | " +
                "[-d|--ignore-directory=<in-module directory path>] | " +
                "[-t|--ignore-tree=<in-module tree root path>] | " +
                "[-r|--ignore-regexp=<in-module path regexp>]]... " +
                "[-p|--prefix=<PREFIX>] " +
                "[-o|--output=<file path>] [-O|--format=<rpm|txt|csv>] " +
                "[-q|--quiet] SOURCE... DESTDIR"

    modVer             := cmd.String(modVerProto)
    modPre             := cmd.String(modPreProto)
    modCommit          := cmd.String(modCommitProto)
    modTime            := cmd.String(modTimeProto)
    modBuild           := cmd.String(modBuildProto)
    modEpoch           := cmd.Int(modEpochProto)
    includeMods        := cmd.Strings(includeModsProto)
    ignoreMods         := cmd.Strings(ignoreModsProto)
    includeFiles       := cmd.Strings(includeFilesProto)
    includeExtensions  := cmd.Strings(includeExtensionsProto)
    includeDirectories := cmd.Strings(includeDirectoriesProto)
    includeTrees       := cmd.Strings(includeTreesProto)
    includeRegexps     := cmd.Strings(includeRegexpsProto)
    ignoreFiles        := cmd.Strings(ignoreFilesProto)
    ignoreExtensions   := cmd.Strings(ignoreExtensionsProto)
    ignoreDirectories  := cmd.Strings(ignoreDirectoriesProto)
    ignoreTrees        := cmd.Strings(ignoreTreesProto)
    ignoreRegexps      := cmd.Strings(ignoreRegexpsProto)
    prefix             := cmd.String(prefixProto)
    output             := cmd.String(outputProto)
    outputFormat       := cmd.String(outputFormatProto)
    quiet              := cmd.Bool(quietProto)
    sources            := cmd.Strings(sourcesProto)
    dstDir             := cmd.String(dstDirProto)

    cmd.Action = func() {
      ver, err := version.New(*modVer, *modPre, *modCommit, *modTime,
                              *modBuild, *modEpoch, *quiet)
      check(err)
      filters      := filter.Default
      filters, err  = filters.AddByType(*includeFiles, *includeExtensions,
                                        *includeDirectories, *includeTrees,
                                        *includeRegexps,
                                        *ignoreFiles,  *ignoreExtensions,
                                        *ignoreDirectories,  *ignoreTrees,
                                        *ignoreRegexps)
      check(err)
      _, err = pack.Scan(*sources, ver, *includeMods, *ignoreMods, filters,
                         *prefix, *dstDir, *output, *outputFormat, *quiet)
      check(err)
    }
  })

  app.Command("index", "Index a GOPROXY Go module repository", func(cmd *cli.Cmd) {

    cmd.LongDesc = `Index a repository of Go modules that conforms to
https://tip.golang.org/cmd/go/#hdr-Module_proxy_protocol`

    cmd.Spec = "DESTDIR"

    dstDir  := cmd.String(dstDirProto)

    cmd.Action = func() {
      _, err := index.Goproxy(*dstDir)
      check(err)
    }
  })

  app.Run(os.Args)
}


