// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package descriptor

import (
  "errors"
  "fmt"
  "io/ioutil"
  "os"
  "path/filepath"
  "regexp"
  "sort"
  "golang.org/x/mod/modfile"
  "golang.org/x/mod/module"
  "golang.org/x/mod/semver"
  "pagure.io/modist/version"
)

var ErrNoDescriptor  = errors.New("no readable module descriptor")
var ErrBadDescriptor = errors.New("no parseable module descriptor")
var ErrBadExt        = errors.New("bad file extension")

type Descriptor struct {
  *modfile.File
}

// Parses a Go .mod module descriptor, without using the “go mod” command
func New(path string) (*Descriptor, error) {
  path = filepath.Clean(path)
  if filepath.Ext(path) != ".mod" {
    return nil, ErrBadExt
  }
  rawdata, err := ioutil.ReadFile(path)
  if err != nil {
    return nil, ErrNoDescriptor
  }
  data, err := modfile.Parse(path, rawdata, nil)
  if err != nil {
    return &Descriptor{data}, ErrBadDescriptor
  }
 return &Descriptor{data}, nil
}

func mod4text(mod module.Version) (result string) {
  result = mod.Path
  if mod.Version != "" {
    result = result + " " + mod.Version
  }
  return result
}

// Remove module directives that make no sense in a clean system component environment
func (descriptor *Descriptor) Trim(quiet bool) error {
  if !quiet {
    fmt.Printf(" – trimming     %s\n", descriptor.Path())
  }
  // Remove local replaces to force the use of the system component version and
  // warn on others
  for _, r := range descriptor.Replace {
    local, err := regexp.MatchString(`^(\.\.?)?/`, r.New.Path)
    if err != nil {
      return fmt.Errorf("could not check if %s is local, %v", r.New.Path, err)
    }
    if local {
      if !quiet {
        fmt.Printf("     removing local replacement of %s, please package it separately\n", mod4text(r.Old))
      }
      if err:= descriptor.DropReplace(r.Old.Path, r.Old.Version) ; err != nil {
        return fmt.Errorf("could not remove the replacement of  %s, %v", mod4text(r.Old), err)
      }
    } else if !quiet {
       fmt.Fprintf(os.Stderr,"   WARNING replacement of %s with %s\n", mod4text(r.Old), mod4text(r.New))
       fmt.Fprintf(os.Stderr,"   WARNING Requires and BuildRequires may differ: please fix import statements\n")
    }
  }

  // Remove indirect requires, the system package manager will resolve them at
  // actual install time
  for _, r := range descriptor.Require {
    if r.Indirect {
      if !quiet {
        fmt.Printf("     removing indirect dependency on %s\n", mod4text(r.Mod))
      }
      if err := descriptor.DropRequire(r.Mod.Path) ; err != nil {
        return fmt.Errorf("could not remove the indirect dependency on %s, %v", r.Mod.Path, err)
      }
    }
  }

  descriptor.Cleanup()
  return nil
}

func (descriptor *Descriptor) Path() string {
  return descriptor.Module.Mod.Path
}

type Provides map[string]*version.Version

// Returns the provides associated with a module descriptor
// To be future-proof, can return multiple results, even though the initial
// implementation only returns one item
func (descriptor *Descriptor) Provides(path string, ver *version.Version,
                                  quiet bool) (results Provides, err error) {
  results = make(Provides)
  mp := descriptor.Path()
  results[mp] = ver
  if results[mp] != nil {
    if err := module.Check(mp, results[mp].Go()) ; err != nil {
      return results, err
    }
  }
  return results, nil
}

func (provides Provides) RPM(onlyName bool, onlyVersion bool, namespace string,
                            ) (results string) {
  var resultlist []string
  var provs []string
  for prov := range provides {
    provs = append(provs, prov)
  }
  sort.Strings(provs)
  for _, prov := range provs {
    line := ""
    sep  := ""
    if !onlyName && provides[prov] != nil {
      line = provides[prov].RPM()
      sep  = " = "
    }
    if !onlyVersion {
      line = namespace + "(" + prov + ")" + sep + line
    }
    resultlist = append(resultlist, line)
  }
  sort.Strings(resultlist)
  for _, result := range resultlist {
     results = results + result + "\n"
  }
  return results
}

func (provides Provides) Go(onlyName bool, onlyVersion bool) (results string) {
  var resultlist []string
  var provs []string
  for prov := range provides {
    provs = append(provs, prov)
  }
  sort.Strings(provs)
  for _, prov := range provs {
    line := ""
    sep  := ""
    if !onlyName && provides[prov] != nil {
      line = provides[prov].Go()
      sep  = " "
    }
    if !onlyVersion {
      line = prov + sep + line
    }
    resultlist = append(resultlist, line)
  }
  sort.Strings(resultlist)
  for _, result := range resultlist {
     results = results + result + "\n"
  }
  return results
}

var ErrUnknownOutputFormat = errors.New("unknown output format")

func (provides Provides) Format(format string, onlyName bool, onlyVersion bool,
                                namespace string) (string, error) {
  switch format {
    case "rpm":
      return provides.RPM(onlyName, onlyVersion, namespace), nil
    case "go":
      return provides.Go(onlyName, onlyVersion),  nil
  }
  return "", ErrUnknownOutputFormat
}

type VersionRange struct {
  Floor      *version.Version `json:",omitempty"`
  Excludes []*version.Version `json:",omitempty"`
}

func (vr *VersionRange) Clean() {
  exs := vr.Excludes
  sort.Slice(exs, func(i, j int) bool {
                    return semver.Compare(exs[i].Go(),
                                          exs[j].Go()) == -1 })
  vr.Excludes = make([]*version.Version, 0, len(exs))
  for i, _ := range exs {
    if semver.Compare(vr.Floor.Go(), exs[i].Go()) == -1 {
      if i == 0 || i > 0 && semver.Compare(exs[i-1].Go(), exs[i].Go()) == -1 {
        vr.Excludes = append(vr.Excludes, exs[i])
      }
    }
  }
}

type Requires map[string]*VersionRange

var ErrDuplicateRequires = errors.New("the required module path has already been declared")

// Returns the provides associated with a module descriptor
func (descriptor *Descriptor) AnyRequires(build bool, quiet bool) (results Requires, err error) {
  results = make(Requires)
  if err = descriptor.Trim(quiet) ; err != nil {
    return results, err
  }
  replaces := make(map[module.Version]module.Version)
  if build {
    for _, replace := range descriptor.Replace {
      replaces[replace.Old] = replace.New
    }
  }
  excludes := make(map[string]map[string]struct{})
  if build {
    for _, exclude := range descriptor.Exclude {
      if _, exists := excludes[exclude.Mod.Path] ; !exists {
        excludes[exclude.Mod.Path] = make(map[string]struct{})
      }
      excludes[exclude.Mod.Path][exclude.Mod.Version] = struct{}{}
    }
  }
  for _, require := range descriptor.Require {
    m := require.Mod
    if r, exists := replaces[m] ; exists {
      m = r
    } else if r, exists := replaces[module.Version{m.Path, ""}] ; exists {
      m = r
    }
    if _, exists := results[m.Path] ; exists {
      return results, ErrDuplicateRequires
    }
    var vr VersionRange
    results[m.Path] = &vr
    vr.Floor, err = version.FromGo(m.Version, quiet)
    if err != nil {
       return results, err
    }
    if _, exists := excludes[m.Path] ; exists {
      for ex := range excludes[m.Path] {
        exclude, err := version.FromGo(ex, quiet)
        if err != nil {
          return results, err
        }
        vr.Excludes = append(vr.Excludes, exclude)
      }
      vr.Clean()
    }
  }
  return results, nil
}

func (descriptor *Descriptor) Requires(quiet bool) (results Requires, err error) {
  return descriptor.AnyRequires(false, quiet)
}

func (descriptor *Descriptor) BuildRequires(quiet bool) (results Requires, err error) {
  return descriptor.AnyRequires(true, quiet)
}

func (requires Requires) RPM(namespace string) (results string) {
  var resultlist []string
  var reqs []string
  for req := range requires {
    reqs = append(reqs, req)
  }
  sort.Strings(reqs)
  for _, req := range reqs {
    rp   := namespace + "(" + req + ")"
    line := rp
    if requires[req].Floor != nil {
      line = line + " >= " + requires[req].Floor.RPM()
    }
    exn := len(requires[req].Excludes)
    if exn > 0 {
      exs := ""
      for i, ex := range requires[req].Excludes {
        exv := ex.RPM()
        if i > 0 {
          exs = exs + " or "
        }
        exs = exs + rp + " = " + exv
      }
      if exn > 1 {
        exs = "(" + exs + ")"
      }
      line = "(" + line + " without " + exs + ")"
    }
    resultlist = append(resultlist, line)
  }
  sort.Strings(resultlist)
  for _, result := range resultlist {
     results = results + result + "\n"
  }
  return results
}

func (requires Requires) Go() (results string) {
  var resultlist []string
  var reqs []string
  for req := range requires {
    reqs = append(reqs, req)
  }
  sort.Strings(reqs)
  for _, req := range reqs {
    line := req
    if requires[req].Floor != nil {
      line = line + " " + requires[req].Floor.Go()
    }
    if len(requires[req].Excludes) > 0 {
      line = line + " excluding:"
      for _, ex := range requires[req].Excludes {
        line = line + " " + ex.Go()
      }
    }
    resultlist = append(resultlist, line)
  }
  sort.Strings(resultlist)
  for _, result := range resultlist {
     results = results + result + "\n"
  }
  return results
}

func (requires Requires) Format(format string, namespace string,
                               ) (string, error) {
  switch format {
    case "rpm":
      return requires.RPM(namespace), nil
    case "go":
      return requires.Go(),  nil
  }
  return "", ErrUnknownOutputFormat
}
