// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package files

import (
  "os"
  "path/filepath"
  "syscall"
  "time"
  "pagure.io/modist/descriptor"
  "pagure.io/modist/modfiles"
  "pagure.io/modist/version"
)

type Files struct {
  // The module root directory fileystem path
  Root      string
  // The module version
  Version   *version.Version
  // The module descriptor
  Descriptor *descriptor.Descriptor `json:"-"`
  // The list of the relative module paths of the files that will be included
  Selected []string                 `json:",omitempty"`
  // The list of the relative module paths of the files ignored after filtering
  Ignored  []string                 `json:",omitempty"`
}

// Converts a modfiles.ModFiles structure to a Files structure, without
// changing path state
func New(module *modfiles.ModFiles) *Files {
  var converted Files
  converted.Root       = module.Root
  converted.Version    = module.Version
  converted.Descriptor = module.Descriptor
  for path, selected := range module.Files {
    if selected {
      converted.Selected = append(converted.Selected, path)
    } else {
      converted.Ignored  = append(converted.Ignored, path)
    }
  }
  return &converted
}

// Convert back when needed
func (module *Files) ModFiles() (*modfiles.ModFiles) {
  var converted modfiles.ModFiles
  converted.Root       = module.Root
  converted.Version    = module.Version
  converted.Descriptor = module.Descriptor
  converted.Files = make(map[string]bool)
  for _, path := range module.Selected {
    converted.Files[path] = true
  }
  for _, path := range module.Ignored {
    converted.Files[path] = false
  }
  return &converted
}

// Returns the most recent modification time of the selected files
// within a module
func (module *Files) MTime() time.Time {
  var t time.Time
  for _, f := range module.Selected {
    p := filepath.Join(module.Root, f)
    if info, err := os.Lstat(p); !os.IsNotExist(err) {
      stat      := info.Sys().(*syscall.Stat_t)
      sec, nsec := stat.Mtim.Unix()
      mtime     := time.Unix(sec, nsec).UTC()
      if mtime.After(t) {
        t = mtime
      }
    }
  }
  return t
}

// Change module timestamp to the module mtime if necessary
func (module *Files) UpdateTimestamp() {
  if module.Version != nil {
    var t time.Time
    if module.Version.Time == t {
      module.Version.Time = module.MTime()
    }
  }
}

func (module *Files) Path() string {
  return module.Descriptor.Path()
}

func (module *Files) PathProvides() (pathprovides descriptor.Provides) {
  pathprovides = make(descriptor.Provides)
  pathprovides[module.Path()] = module.Version
  for _, f := range module.Selected {
    d := filepath.Join(module.Path(), filepath.Dir(f))
    pathprovides[d] = module.Version
  }
  return pathprovides
}
