// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package filter

import (
  "encoding/json"
  "errors"
  "fmt"
  "path/filepath"
  "regexp"
  "sort"
  "strings"
  "time"
  "pagure.io/modist/descriptor"
  "pagure.io/modist/files"
  "pagure.io/modist/list"
  "pagure.io/modist/modfiles"
  "pagure.io/modist/version"
  "pagure.io/modist/write"
)

// A module content filter
type Filter struct {
  Match  func(path string) bool
  Select bool // Select or ignore paths that match
}

type Filters []Filter

var ErrUnknownFilter = errors.New("unknown filter kind")

func New(include bool, kind string, expr string) (Filter, error) {
  var filter Filter
  filter.Select = include
  switch kind {
    case "file":
      filter.Match = func(path string) bool {
        return path == filepath.Clean(expr)
      }
    case "extension":
      filter.Match = func(path string) bool {
        return filepath.Ext(path) == "." + strings.TrimPrefix(expr,".")
      }
    case "directory":
      filter.Match = func(path string) bool {
        return filepath.Dir(path) == filepath.Clean(expr)
      }
    case "tree":
      filter.Match = func(path string) bool {
        rel, err := filepath.Rel(expr, path)
        return err == nil && !strings.HasPrefix(rel, "../")
      }
    case "regexp":
      re, err := regexp.Compile(expr)
      if err != nil {
        return filter, err
      }
      filter.Match = func(path string) bool {
        return re.Match([]byte(path))
      }
    default:
      return filter, ErrUnknownFilter
  }
  return filter, nil
}

func (filters Filters) Add(include bool, kind string, exprs []string) (results Filters, err error) {
  results = filters
  for _, expr := range exprs {
    filter, err := New(include, kind, expr)
    if err != nil {
      return results, err
    }
    results = append(results, filter)
  }
  return results, nil
}

func (filters Filters) AddByType(includeFiles []string, includeExtensions []string, includeDirectories []string,
                                 includeTrees []string, includeRegexps []string,
                                 ignoreFiles []string,  ignoreExtensions []string,  ignoreDirectories []string,
                                 ignoreTrees []string,  ignoreRegexps []string) (results Filters, err error) {
  results, err  = filters.Add(true, "file",       includeFiles)
  if err != nil {
    return results, err
  }
  results, err  = results.Add(true, "extension",  includeExtensions)
  if err != nil {
    return results, err
  }
  results, err  = results.Add(true, "directory",  includeDirectories)
  if err != nil {
    return results, err
  }
  results, err  = results.Add(true, "tree",       includeTrees)
  if err != nil {
    return results, err
  }
  results, err  = results.Add(true, "regexp",     includeRegexps)
  if err != nil {
    return results, err
  }
  results, err  = results.Add(false, "file",      ignoreExtensions)
  if err != nil {
    return results, err
  }
  results, err  = results.Add(false, "extension", ignoreExtensions)
  if err != nil {
    return results, err
  }
  results, err  = results.Add(false, "directory", ignoreDirectories)
  if err != nil {
    return results, err
  }
  results, err  = results.Add(false, "tree",      ignoreTrees)
  if err != nil {
    return results, err
  }
  results, err  = results.Add(false, "regexp",    ignoreRegexps)
  return results, err
}

var Default Filters

func init() {
  var filter Filter
  // Include testdata by default since Go modules include tests by default
  filter, _ = New(true, "regexp", "^(.*/)?testdata/")
  Default   = append(Default, filter)
  // Include protobuf files for now, till we get a default filesystem location
  // and get go code to read it
  filter, _ = New(true, "extension", "proto")
  Default   = append(Default, filter)
  // Include markdown files by default
  filter, _ = New(true, "extension", "md")
  Default   = append(Default, filter)
  //Include glide files by default. This may be problematic as we do not clean
  //them like we clean Go descriptor files
  filter, _ = New(true, "file", "glide.yaml")
  Default   = append(Default, filter)
  filter, _ = New(true, "file", "glide.lock")
  Default   = append(Default, filter)
  // Ignore anything that looks like example code as upstream example code
  // quality is terrible
  filter, _ = New(false, "regexp", "example")
  Default  = append(Default, filter)
}

// Applies filters to a modfiles.ModFiles structure, changing the input structure
func Module(module *modfiles.ModFiles, filters []Filter) *files.Files {
  index := files.New(module)
  for _, filter := range filters {
    if filter.Select {
      for _, path := range index.Ignored {
        if filter.Match(path) {
          module.Files[path] = true
        }
      }
    } else {
      for _, path := range index.Selected {
        if filter.Match(path) {
          module.Files[path] = false
        }
      }
    }
  }
  result := files.New(module)
  sort.Strings(result.Selected)
  sort.Strings(result.Ignored)
  return result
}

type Files map[string]*files.Files

func (modules *Files) Text(selected bool) string {
  var modPaths []string
  for modPath := range *modules {
    modPaths = append(modPaths, modPath)
  }
  sort.Strings(modPaths)
  var data string
  for _, modPath := range modPaths {
    modFiles := (*modules)[modPath]
    if selected {
      if modFiles.Selected != nil {
        data = data + fmt.Sprintf(" – retained in  %s\n", modPath)
        for _, file := range modFiles.Selected {
          data = data + "     " + file + "\n"
        }
      }
    }
    if modFiles.Ignored != nil {
      data = data + fmt.Sprintf(" – ignored in  %s\n", modPath)
      for _, file := range modFiles.Ignored {
        data = data + "     " + file + "\n"
      }
    }
  }
  return data
}

func (modules *Files) JSON() (results string, err error) {
  data, err := json.MarshalIndent(modules, "", "  ")
  if err != nil {
    return "", fmt.Errorf("failed to marchal results as json data,  %v", err)
  }
  return string(data[:]), nil
}

var ErrUnknownOutputFormat = errors.New("unknown output format")

func (modules *Files) Format(format string, selected bool) (string, error) {
  switch format {
    case "txt":
      return modules.Text(selected), nil
    case "json":
      return modules.JSON()
  }
  return "", ErrUnknownOutputFormat
}

// List the modules found in the tree located at root, applying filters to
// their file lists, reporting the result in output with outputFormat while
// being quiet
func Scan(sources []string, ver *version.Version,
          includeMods []string, ignoreMods []string, filters []Filter,
          quiet bool) (files Files, err error) {
  rawModules, err := list.Scan(sources, ver, includeMods, ignoreMods, false,
                               quiet)
  if err != nil {
    return nil, err
  }
  var modules []string
  rawModules.RLock()
  for module := range rawModules.Modules {
    modules = append(modules, module)
  }
  sort.Strings(modules)
  files = make(Files)
  for _, module := range modules {
    if !quiet {
      fmt.Printf(" – filtering    %s\n", module)
    }
    files[module] = Module(rawModules.Modules[module], filters)
    if !quiet {
      fmt.Printf(" – timestamping %s\n", module)
    }
    files[module].UpdateTimestamp()
    if !quiet {
      fmt.Printf("     %s\n",
                 files[module].Version.Time.UTC().Format(time.RFC3339))
      fmt.Printf("     %s\n",
                 files[module].Version.Go())
    }
  }
  rawModules.RUnlock()
  return files, nil
}

func List(sources []string, ver *version.Version,
          includeMods []string, ignoreMods []string, filters []Filter,
          output string, outputFormat string, selected bool, quiet bool,
         ) (files Files,  written write.Files, err error) {
  written = make(write.Files)
  files, err = Scan(sources, ver, includeMods, ignoreMods, filters, quiet)
  if err != nil {
    return files, written, err
  }
  report, err := files.Format(outputFormat, selected)
  if err != nil {
    return files, written, err
  }
  if !quiet {
    fmt.Printf("\nResults:\n")
  }
  if report != "" {
    fmt.Printf("%s", report)
  }
  if output != "" {
    err = written.File(strings.NewReader(report), "", output, "txt", "", true)
  }
  return files, written, err
}

func (modules Files) PathProvides() (results descriptor.Provides) {
  results = make(descriptor.Provides)
  for _, mod := range modules {
    for p, v := range mod.PathProvides() {
      results[p] = v
    }
  }
  return results
}

func (modules Files) IPathProvides(quiet bool) (results descriptor.Provides,
                                                err error) {
  results = make(descriptor.Provides)
  for _, mod := range modules {
    r, err := mod.ModFiles().Provides(quiet)
    if err != nil {
      return results, err
    }
    for p, v := range r {
      results[p] = v
    }
  }
  return results, nil
}

func PathProvides(sources []string, ver *version.Version,
                  includeMods []string, ignoreMods []string, filters []Filter,
                  onlyName bool, onlyVersion bool,
                  output string, outputFormat string, quiet bool,
                 ) (files Files, written write.Files, err error) {
  written = make(write.Files)
  files, err = Scan(sources, ver, includeMods, ignoreMods, filters, quiet)
  if err != nil {
    return files, written, err
  }
  ipathprovides, err := files.IPathProvides(quiet)
  if err != nil {
    return files, written, err
  }
  report1, err  := ipathprovides.Format(outputFormat, onlyName, onlyVersion,
                                        "golang-ipath")
  if err != nil {
    return files, written, err
  }
  pathprovides  := files.PathProvides()
  report2, err := pathprovides.Format(outputFormat, onlyName, onlyVersion, "golang")
  report := report1 + report2
  if err != nil {
    return files, written, err
  }
  if !quiet {
    fmt.Printf("\nResults:\n")
  }
  if report != "" {
    fmt.Printf("%s", report)
  }
  if output != "" {
    err = written.File(strings.NewReader(report), "", output, "txt", "", true)
  }
  return files, written, err
}

