// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package index

import (
  "errors"
  "io/ioutil"
  "os"
  "path/filepath"
  "sort"
  "strings"
  "golang.org/x/mod/module"
  "golang.org/x/mod/semver"
  "pagure.io/modist/descriptor"
  "pagure.io/modist/read"
  "pagure.io/modist/write"
)

var ErrNotAModuleDirectory  = errors.New("the target is not a valid module directory")
var ErrInvalidModuleVersion = errors.New("mismatch between the version used in the module filename, the version written in its info file, and GOPROXY encoding rules")
var ErrNothingToIndex       = errors.New("no module to index here, the directory was removed")

// Loosely modeled after rewriteVersionList in cmd/go/internal/modfetch/cache.go
// which is not exported anywhere so far
func Directory(dir string, written write.Files) (err error) {
  if filepath.Base(dir) != "@v" {
    return ErrNotAModuleDirectory
  }

  listFile := filepath.Join(dir, "list")
  // Without the rest of the internal go code, no nice locking logic
  infos, err := ioutil.ReadDir(dir)
  if err != nil {
    return
  }
  var list []string
  for _, info := range infos {
    name    := info.Name()
    radical := strings.TrimSuffix(name, ".mod")
    if radical != name {
      // Test if this is a valid module descriptor
      _, err  := descriptor.New(filepath.Join(dir, radical + ".mod"))
      if err != nil {
        return err
      }
      // Check the corresponding info file
      rev, err := read.Info(filepath.Join(dir, radical + ".info"))
      if err != nil {
        return err
      }
      if module.CanonicalVersion(rev.Version) != rev.Version {
        return ErrInvalidModuleVersion
      }
      encVer, err := module.EscapeVersion(rev.Version)
      if encVer != radical {
        return ErrInvalidModuleVersion
      }
      // If we were even more paranoïd we'd also check zip payload hashes
      list = append(list, rev.Version)
    }
  }
  if len(list) == 0 {
    if err := os.Remove(listFile) ; err != nil {
      return err
    }
    if err := os.Remove(dir) ; err != nil {
      return err
    }
    return ErrNothingToIndex
  }
  sort.Slice(list, func(i, j int) bool {
                     return semver.Compare(list[i], list[j]) == -1 })
  var data string
  for _, ver := range list {
    encVer, _ := module.EscapeVersion(ver)
    data = data + encVer + "\n"
  }
  return written.File(strings.NewReader(data), "", listFile, "", "", true)
}

func goproxyWalker(results write.Files, path string, info os.FileInfo,
                   err error) error {
  if err != nil {
    return err
  }
  if info.IsDir() {
    err := Directory(path, results)
    if err == ErrNothingToIndex {
      return filepath.SkipDir
    }
    if err != nil && err != ErrNotAModuleDirectory {
      return err
    }
  }
  return nil
}

func Goproxy(path string) (results write.Files, err error) {
  results = make(write.Files)
  err = filepath.Walk(path,
                      func(p string, info os.FileInfo, err error) error {
                        return goproxyWalker(results, p, info, err)
                      })
  return results, err
}
