// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package install

import (
  "fmt"
  "os"
  "path/filepath"
  "sort"
  "pagure.io/modist/files"
  "pagure.io/modist/filter"
  "pagure.io/modist/read"
  "pagure.io/modist/version"
  "pagure.io/modist/write"
)

// Install a module to destination
func Module(module *files.Files,
            prefix string, qualifyRoot bool, destination string,
            written write.Files, quiet bool) error {
  // Create the target directory
  if err := os.MkdirAll(filepath.Join(prefix, destination), 0777) ;
     err != nil {
    return err
  }
  modPath := module.Descriptor.Module.Mod.Path
  if qualifyRoot {
    modPath += "@" + module.Version.Go()
  }
  var mdirs []string
  for root := modPath ; root != "." ; root = filepath.Dir(root) {
    mdirs = append (mdirs, filepath.Join(destination, root))
  }
  sort.Strings(mdirs)
  for _, mdir := range mdirs {
  if err := written.Dir(prefix, mdir, "", true) ; err != nil {
      return err
    }
  }
  if err := written.Descriptor(module.Descriptor, prefix,
            filepath.Join(destination, modPath, "go"), "", quiet) ; err != nil {
    return err
  }
  if err := written.Info(module.Version, prefix,
            filepath.Join(destination, modPath, "go"), "") ; err != nil {
    return err
  }
  var dirs []string
  dirmap := make(map[string]bool)
  for _, path := range module.Selected {
    dirmap[filepath.Dir(path)] = true
  }
  delete(dirmap, ".")
  for dir := range dirmap {
    for root := filepath.Dir(dir) ; root != "." ; root = filepath.Dir(root) {
      dirmap[root] = true
    }
  }
  for dir := range dirmap {
   dirs = append (dirs, dir)
  }
  sort.Strings(dirs)
  for _, dir := range dirs {
    err := written.Dir(prefix, filepath.Join(destination, modPath, dir), "", quiet)
    if err != nil {
      return err
    }
  }
  paths := make(map[string]bool)
  for _, path := range module.Selected {
    paths[path] = true
  }
  for _, path := range module.Selected {
    spath := filepath.Join(module.Root, path)
    mpath := filepath.Join(destination, modPath)
    dpath := filepath.Join(mpath, path)
    target, mt, st, err := read.Symlink(module.Root, path)
    switch err {
      case read.ErrNotASymLink:
        var file *os.File
        file, err = os.Open(spath)
        if err != nil {
          return err
        }
        err = written.File(file, prefix, dpath, "", spath, false)
      case read.ErrOutsideSymlink:
        fmt.Fprintf(os.Stderr,"   WARNING %s target %s is outside %s\n", dpath, target, mpath)
        err = written.Symlink(prefix, dpath, target, spath)
      case nil:
        if !quiet && target != st {
          fmt.Printf("     simplifying %s target %s to %s\n", dpath, target, st)
        }
        _, existsP := paths[mt]
        _, existsD := dirmap[mt]
        if !existsP && !existsD {
          fmt.Fprintf(os.Stderr,"   WARNING %s target %s is not included in the module\n", path, st)
        }
        err = written.Symlink(prefix, dpath, st, spath)
    }
    if err != nil {
      return err
    }
  }
  // No use and confusing to do the same for mdirs as they may be shared
  sort.Sort(sort.Reverse(sort.StringSlice(dirs)))
  for _, dir := range dirs {
    if err := write.SetModTime(filepath.Join(prefix, destination, modPath, dir), filepath.Join(module.Root, dir)) ; err != nil {
      return err
    }
  }
  return nil
}

// Install the modules found in the tree located at root, into destination,
// prefixed by prefix, applying filters to their file lists, reporting the
// result in output with outputFormat and being quiet
func Scan(sources []string, ver *version.Version,
          includeMods []string, ignoreMods []string, filters []filter.Filter,
          prefix string, qualifyRoot bool, destination string,
          output string, outputFormat string, quiet bool,
         ) (written write.Files, err error) {
  written  = make(write.Files)
  rawModules, err := filter.Scan(sources, ver, includeMods, ignoreMods, filters,
                                 quiet)
  if err != nil {
    return written, err
  }
  var modules []string
  for module := range rawModules {
    modules = append(modules, module)
  }
  sort.Strings(modules)
  for _, module := range modules {
    if !quiet {
      fmt.Printf(" – installing   %s\n", module)
    }
    err = Module(rawModules[module], prefix, qualifyRoot, destination, written,
                 quiet)
    if err != nil {
      return written, err
    }
  }
  if !quiet {
    report, err := written.Format(outputFormat)
    if err != nil {
      return written, err
    }
    fmt.Printf("\nResults:\n%s", report)
  }
  if output != "" {
    err = written.Output(output, outputFormat)
  }
  return written, err
}
