// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package list

import (
  "encoding/json"
  "errors"
  "fmt"
  "os"
  "path/filepath"
  "sort"
  "strings"
  "sync"
  "golang.org/x/mod/module"
  "golang.org/x/mod/semver"
  "pagure.io/modist/descriptor"
  "pagure.io/modist/modfiles"
  "pagure.io/modist/read"
  "pagure.io/modist/version"
  "pagure.io/modist/write"
)

type Results struct{
  sync.RWMutex
  Modules map[string]*modfiles.ModFiles
}

// Processes a file (path) within a module tree
func moduleWalker(modPath string, ver *version.Version, addUnknownModules bool,
                  modules map[string]bool, results *Results, onlydescr bool,
                  sourcefiles bool, detachsubs bool, walksubs bool, quiet bool,
                  path string, info os.FileInfo, err error) error {
  if err != nil {
    return err
  }
  results.RLock()
  relpath, err := filepath.Rel(results.Modules[modPath].Root, path)
  results.RUnlock()
  if err != nil {
    return err
  }
  if info.IsDir() {
    if relpath != "." {
      if detachsubs {
        info, err := TryModule(filepath.Join(path, "go.mod"), quiet)
        if err == nil { // Another module here
          if !walksubs { // But we don’t care
            return filepath.SkipDir
          }
          err := Descriptor(info, ver, addUnknownModules, modules, results,
                            onlydescr, walksubs, quiet)
          if err == ErrSeenModule { // We already know about this tree
            return filepath.SkipDir
          }
          if err == ErrIgnoredModule { // Let’s pretend the tree is something
                                       // not a module and not part of the
                                       // current module
            err = filepath.Walk(path,
                                func(p string, info os.FileInfo, err error,
                                    ) error {
                                  return treeWalker(path, ver,
                                                    addUnknownModules, modules,
                                                    results, onlydescr, quiet,
                                                    p, info, err)
                                })
          }
          if err == nil { // The directory has been succesfully analyzed
            return filepath.SkipDir
          }
        }
        if err != descriptor.ErrNoDescriptor {
          return err
        }
      }
      dir := filepath.Base(relpath)
      if dir == "testdata" || dir == "vendor" || strings.HasPrefix(dir, ".") ||
                strings.HasPrefix(dir, "_") {
        if onlydescr {
          return filepath.SkipDir
        }
        if sourcefiles {
          err =  filepath.Walk(path,
                               func(p string, info os.FileInfo, err error,
                                   ) error {
                                 return moduleWalker(modPath, ver,
                                                     addUnknownModules, modules,
                                                     results, false, false,
                                                     false, false, quiet, p,
                                                     info, err)
                               })
          if err == nil {
            return filepath.SkipDir
          }
          return err
        }
      }
      if sourcefiles {
        results.Lock()
        err = results.Modules[modPath].Sources(path)
        results.Unlock()
        return err
      }
    }
  } else { // Just a normal file, add it as potential module file
    if !onlydescr && relpath != "go.mod" {
      // The module descriptor always exists but needs special care
      results.RLock()
      if _, exists := results.Modules[modPath].Files[relpath] ; !exists {
        results.RUnlock()
        results.Lock()
        if _, exists := results.Modules[modPath].Files[relpath] ; !exists {
          results.Modules[modPath].Files[relpath] = false
        }
        results.Unlock()
      } else {
        results.RUnlock()
      }
    }
  }
  return nil
}

// Taking only module into account: we do not allow several versions of the
// same module path within a single source directory, because a source
// directory is associated with a single project version
var ErrSeenModule      = errors.New("this module was already scanned")
var ErrDuplicateModule = errors.New(
  "this module conflicts with another one that declares the same path")
var ErrIgnoredModule   = errors.New(
  "the user requested to ignore this module path")

// Must be called with results RLock or Lock held
func checkSeenModule(modPath string, path string, results *Results) error {
  other, exists := results.Modules[modPath]
  if exists {
    otherinfo, err := os.Stat(other.Root)
    if err != nil {
      return err
    }
    info, err      := os.Stat(path)
    if err != nil {
      return err
    }
    if os.SameFile(otherinfo, info) {
      fmt.Fprintf(os.Stderr,"   WARNING already seen %s at %s\n", modPath, path)
      return ErrSeenModule
    }
    return ErrDuplicateModule
  }
  return nil
}

// Check if a file path is a valid module descriptor; return a
// modfiles.ModFiles struct if true. The Files part of the result struct is not
// populated and needs separate walking
func TryModule(path string, quiet bool) (result modfiles.ModFiles, err error) {
  descr, err := descriptor.New(path)
  if err != nil {
    return result, err
  }
  result.Root        = filepath.Dir(path)
  result.Descriptor  = descr
  result.Files       = make(map[string]bool)
  // We don’t care if reading the Version fails – Version info may be missing
  // in the first place
  result.Version, _  = read.Version(path, quiet)
  return result, err
}

// Walks a module tree to list its constituant files,
func Module(root string, ver *version.Version, addUnknownModules bool,
            modules map[string]bool, results *Results,
            onlydescr bool, quiet bool) error {
  info, err := TryModule(filepath.Join(root, "go.mod"), quiet)
  if err != nil {
   return err
  }
  return Descriptor(info, ver, addUnknownModules, modules, results,
                    onlydescr, true, quiet)
}

// Walks a module tree, identified by its modfiles.ModFiles info
func Descriptor(info modfiles.ModFiles, ver *version.Version,
                addUnknownModules bool, modules map[string]bool,
                results *Results, onlydescr bool, walksubs bool, quiet bool,
               ) (err error) {
  modPath  := info.Path()
  root     := info.Root
  _, known := modules[modPath]
  if !((!known && addUnknownModules) || (known && modules[modPath])) {
    if !quiet {
      fmt.Printf(" – ignoring     %s at %s\n", modPath, root)
    }
    return ErrIgnoredModule
  }
  if !quiet {
    fmt.Printf(  " – listing      %s at %s\n", modPath, root)
  }
  results.RLock()
  err = checkSeenModule(modPath, root, results)
  results.RUnlock()
  if err != nil {
    return err
  }
  results.Lock()
  err = checkSeenModule(modPath, root, results)
  if err != nil {
    results.Unlock()
    return err
  }
  results.Modules[modPath] = &info
  if onlydescr && !walksubs {
    results.Unlock()
    return nil
  }
  if !onlydescr {
    err = info.Sources(root)
    if err != nil {
      results.Unlock()
      return err
    }
  }
  if ver != nil && *ver != (version.Version{}) {
    if err := module.Check(modPath, ver.Go()) ; err != nil {
      return err
    }
    results.Modules[modPath].Version = ver
  }
  results.Unlock()
  return filepath.Walk(root,
    func(path string, info os.FileInfo, err error) error {
      return moduleWalker(modPath, ver, addUnknownModules, modules, results,
                          onlydescr, true, true, walksubs, quiet, path, info,
                          err)
    })
}

// Process a file (path) within a source tree
func treeWalker(root string, ver *version.Version, addUnknownModules bool,
                modules map[string]bool, results *Results, onlydescr bool,
                quiet bool, path string, info os.FileInfo, err error) error {
  if err != nil {
    return err
  }
  if info.IsDir() {
    if root != path {
      err := Module(path, ver, addUnknownModules, modules, results, onlydescr,
                    quiet)
      if err == nil || err == ErrSeenModule { // This a module tree root, and
                                              // it has been analyzed
        return filepath.SkipDir
      } else if (err != descriptor.ErrNoDescriptor) && (err != ErrIgnoredModule) {
        return err
      }
    }
    dir := filepath.Base(path)
    if dir == "testdata" || dir == "vendor" || strings.HasPrefix(dir, ".") ||
              strings.HasPrefix(dir, "_") {
      return filepath.SkipDir
    }
  }
  return nil
}

func selectSources(sources []string, quiet bool) (roots []string,
                   infos []modfiles.ModFiles, err error) {
  srcInfos := make(map[os.FileInfo]string)
  sort.Strings(sources)
  for _, src := range sources {
    srcInfo, err := os.Stat(src)
    if err != nil {
      if !quiet {
        fmt.Printf(" – ignoring source %s (can not stat)\n", src)
       }
       break
    }
    _, exists := srcInfos[srcInfo]
    if !exists {
      for other := range srcInfos {
        if os.SameFile(srcInfo, other) {
          exists = true
          break
        }
      }
    }
    if exists {
      if !quiet {
        fmt.Printf(" – ignoring source %s (already specified)\n", src)
      }
    } else {
      srcInfos[srcInfo] = src
      if srcInfo.IsDir() {
        roots = append(roots, src)
      } else {
        info, err := TryModule(src, quiet)
        if err != nil {
          return roots, infos, err
        }
        infos = append(infos, info)
      }
    }
  }
  return roots, infos, nil
}

func selectModules(includeMods []string, ignoreMods []string,
                  ) (addUnknownModules bool, modules map[string]bool) {
  addUnknownModules = true
  modules = make(map[string]bool)
  if includeMods != nil {
    addUnknownModules = false
    for _, modPath := range includeMods {
      modules[modPath] = true
    }
  }
  for _, modPath := range ignoreMods {
    modules[modPath] = false
  }
  return addUnknownModules, modules
}

// Walks sources to find and read Go module source trees within
func Scan(sources []string, ver *version.Version,
          includeMods []string, ignoreMods []string,
          onlydescr bool, quiet bool,
         ) (results *Results, err error) {
  results    = &Results{
    Modules: make(map[string]*modfiles.ModFiles),
  }
  roots, infos, err := selectSources(sources, quiet)
  if err != nil {
    return results, err
  }
  addUnknownModules, modules := selectModules(includeMods, ignoreMods)
  errors := make(chan error)
  for _, root := range roots {
    go func(root string) {
      if !quiet {
        fmt.Printf("Searching for unpacked Go modules in %s…\n", root)
      }
      err = Module(root, ver, addUnknownModules, modules, results, onlydescr,
                   quiet)
      if (err == descriptor.ErrNoDescriptor) ||
         (err == ErrIgnoredModule)  { // No useful module root here, look deeper
        err = filepath.Walk(root,
                            func(path string, info os.FileInfo, err error,
                                ) error {
                              return treeWalker(root, ver, addUnknownModules,
                                                modules, results, onlydescr,
                                                quiet, path, info, err)
        })
      }
      errors <- err
    }(root)
  }
  for _, info := range infos {
    go func(info modfiles.ModFiles) {
      if !quiet {
        fmt.Printf("Looking at the %s module in %s…\n", info.Path(), info.Root)
      }
      errors <- Descriptor(info, ver, addUnknownModules, modules, results,
                           onlydescr, false, quiet)
    }(info)
  }
  for i := 0 ; i < len(roots)+len(infos) ; i++ {
    if err := <-errors ; err != nil && err != ErrSeenModule {
      return results, err
    }
  }
  return results, nil
}

func Discover(sources []string,
              includeMods []string, ignoreMods []string, onlyRoot bool,
              output string, outputFormat string, quiet bool,
             ) (results *Results, written write.Files, err error) {
  written = make(write.Files)
  results, err = Scan(sources, nil, includeMods, ignoreMods, true, quiet)
  if err != nil {
    return results, written, err
  }
  report, err := results.Format(outputFormat, onlyRoot)
  if err != nil {
    return results, written, err
  }
  if !quiet {
    fmt.Printf("\nResults:\n")
  }
  if report != "" {
    fmt.Printf("%s", report)
  }
  if output != "" {
    err = written.File(strings.NewReader(report), "", output, outputFormat, "",
                       true)
  }
  return results, written, err
}

func (modules *Results) List(separator string, onlyRoot bool,
                            ) (results string) {
  modules.RLock()
  var resultlist []string
  for _, mod := range modules.Modules {
    result      := mod.Root
    if !onlyRoot {
      result = mod.Path() + separator +
               mod.Version.Go() + separator + result
    }
    resultlist   = append(resultlist, result)
  }
  modules.RUnlock()
  sort.Strings(resultlist)
  for _, result := range resultlist {
     results = results + result + "\n"
  }
  return results
}

func (modules *Results) Text(onlyRoot bool) string {
  return modules.List("\t", onlyRoot)
}

func (modules *Results) CSV(onlyRoot bool)  string {
  return modules.List(";", onlyRoot)
}

func (modules *Results) JSON() (string, error) {
  type M struct{
    Module  string
    Version string
  }
  results := make(map[string]M)
  for _, mod := range modules.Modules {
    var m M
    m.Module  = mod.Path()
    m.Version = mod.Version.Go()
    results[mod.Root] = m
  }
  data, err := json.MarshalIndent(results, "", "  ")
  if err != nil {
     return "", fmt.Errorf("failed to marchal results as json data,  %v", err)
  }
  return string(data[:]), nil
}

var ErrUnknownOutputFormat = errors.New("unknown output format")

func (modules *Results) Format(format string, onlyRoot bool) (string, error) {
  switch format {
    case "txt":
      return modules.Text(onlyRoot), nil
    case "csv":
      return modules.CSV(onlyRoot),  nil
    case "json":
      return modules.JSON()
  }
  return "", ErrUnknownOutputFormat
}

var ErrDuplicateProvides = errors.New(
  "the provided module path has already been declared")

func (modules *Results) Provides(quiet bool) (results descriptor.Provides,
                                              err error) {
  results = make(descriptor.Provides)
  modules.RLock()
  for _, mod := range modules.Modules {
    ps, err := mod.Provides(quiet)
    if err != nil {
      return results, err
    }
    for p := range ps {
      if _, exists := results[p] ; exists {
                     // ↓ Should not happen since list rejects them
        return results, ErrDuplicateProvides
      }
      results[p] = ps[p]
    }
  }
  modules.RUnlock()
  return results, err
}

func (modules *Results) anyRequires(build bool, quiet bool,
                                   ) (results descriptor.Requires, err error) {
  results = make(descriptor.Requires)
  var reqs []descriptor.Requires
  modules.RLock()
  for _, mod := range modules.Modules {
    rs, err := mod.AnyRequires(build, quiet)
    if err != nil {
      return results, err
    }
    reqs = append(reqs, rs)
  }
  modules.RUnlock()
  merged := make(map[string]struct{})
  for _, rs := range reqs {
    for r := range rs {
      _, exists := results[r]
      if !exists {
        results[r] = rs[r]
      } else {
        merged[r] = struct{}{}
        if semver.Compare(results[r].Floor.Go(), rs[r].Floor.Go()) == -1 {
          results[r].Floor = rs[r].Floor
        }
        results[r].Excludes = append(results[r].Excludes, rs[r].Excludes...)
      }
    }
  }
  for r := range merged {
    results[r].Clean()
  }
  return results, nil
}

func (modules *Results) Requires(quiet bool) (descriptor.Requires, error) {
  return modules.anyRequires(false, quiet)
}

func (modules *Results) BuildRequires(quiet bool,
                                     ) (results descriptor.Requires,
                                        err error) {
  results = make(descriptor.Requires)
  rawbrs, err := modules.anyRequires(true, quiet)
  if err != nil {
    return results, err
  }
  provs, err := modules.Provides(quiet)
  if err != nil {
    return results, err
  }
  if len(provs) == 1 { // No need to consolidate for a single module
    return rawbrs, nil
  }
  rawbrl := make([]string, 0, len(rawbrs))
  for br := range rawbrs {
    rawbrl = append(rawbrl, br)
  }
  sort.Strings(rawbrl)
  if !quiet {
    fmt.Printf("Consolidating BuildRequires…\n")
  }
  for _, br := range rawbrl {
    retain := true
    _, exists := provs[br]
    if exists {
      refv := "v.0.0"
      if provs[br] != nil {
        refv = provs[br].Go()
      }
      retain = false
      if semver.Compare(refv, rawbrs[br].Floor.Go()) == -1 {
        retain = true
      } else {
        for _, ex := range rawbrs[br].Excludes {
          if semver.Compare(refv, ex.Go()) == 0 {
            retain = true
            break
          }
        }
      }
      if retain {
        fmt.Fprintf(os.Stderr,"   WARNING the provided %s module does not satisfy the build requirements or other modules\n", br)
      }
    }
    if retain {
      results[br] = rawbrs[br]
    }
  }
  return results, nil
}

func Provides(sources []string, ver *version.Version,
              includeMods []string, ignoreMods []string,
              onlyName bool, onlyVersion bool,
              output string, outputFormat string, quiet bool,
             ) (results *Results, written write.Files, err error) {
  written = make(write.Files)
  scan, err := Scan(sources, ver, includeMods, ignoreMods, true, quiet)
  if err != nil {
    return results, written, err
  }
  provides, err := scan.Provides(quiet)
  if err != nil {
    return results, written, err
  }
  report, err := provides.Format(outputFormat, onlyName, onlyVersion, "golang-module")
  if err != nil {
    return results, written, err
  }
  if !quiet {
    fmt.Printf("\nResults:\n")
  }
  if report != "" {
    fmt.Printf("%s", report)
  }
  if output != "" {
    err = written.File(strings.NewReader(report), "", output, "txt", "", true)
  }
  return results, written, err
}

func anyRequires(sources []string, includeMods []string, ignoreMods []string,
                 namespace string, output string, outputFormat string,
                 quiet bool) (results descriptor.Requires, written write.Files,
                              err error) {
  written = make(write.Files)
  scan, err := Scan(sources, nil, includeMods, ignoreMods, true, quiet)
  if err != nil {
    return results, written, err
  }
  requires, err := scan.Requires(quiet)
  if err != nil {
    return results, written, err
  }
  report, err := requires.Format(outputFormat, namespace)
  if err != nil {
    return results, written, err
  }
  if !quiet {
    fmt.Printf("\nResults:\n")
  }
  if report != "" {
    fmt.Printf("%s", report)
  }
  if output != "" {
    err = written.File(strings.NewReader(report), "", output, "txt", "", true)
  }
  return results, written, err
}

func Requires(sources []string, includeMods []string, ignoreMods []string,
	      output string, outputFormat string, quiet bool,
             ) (results descriptor.Requires, written write.Files, err error) {
  return anyRequires(sources, includeMods, ignoreMods, "golang-module", output,
                     outputFormat, quiet)
}

func PathRequires(sources []string, includeMods []string, ignoreMods []string,
                  output string, outputFormat string, quiet bool,
                 ) (results descriptor.Requires, written write.Files, err error) {
  return anyRequires(sources, includeMods, ignoreMods, "golang-ipath", output,
                     outputFormat, quiet)
}

func BuildRequires(sources []string, ver *version.Version,
                   includeMods []string, ignoreMods []string,
                   output string, outputFormat string, quiet bool,
                  ) (results descriptor.Requires, written write.Files,
                     err error) {
  written = make(write.Files)
  scan, err := Scan(sources, ver, includeMods, ignoreMods, true, quiet)
  if err != nil {
    return results, written, err
  }
  buildrequires, err := scan.BuildRequires(quiet)
  if err != nil {
    return results, written, err
  }
  report, err := buildrequires.Format(outputFormat, "golang-module")
  if err != nil {
    return results, written, err
  }
  if !quiet {
    fmt.Printf("\nResults:\n")
  }
  if report != "" {
    fmt.Printf("%s", report)
  }
  if output != "" {
    err = written.File(strings.NewReader(report), "", output, "txt", "", true)
  }
  return results, written, err
}
