// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package modfiles

import (
  "go/build"
  "path/filepath"
  "pagure.io/modist/descriptor"
  "pagure.io/modist/version"
)

// A structure holding raw unpacked module file information
type ModFiles struct {
  // The module descriptor
  Descriptor *descriptor.Descriptor
  // The module version
  Version    *version.Version
  // The module root directory filesystem path
  Root       string
  // A map of the relative paths of the module files within the module, to
  // whether they will be included in the packed module or not.
  Files      map[string]bool
}

func (module *ModFiles) Path() string {
  return module.Descriptor.Path()
}

func (module *ModFiles) Provides(quiet bool) (descriptor.Provides, error) {
  return module.Descriptor.Provides(module.Root, module.Version, quiet)
}

func (module *ModFiles) AnyRequires(build bool,
                                    quiet bool) (descriptor.Requires, error) {
  return module.Descriptor.AnyRequires(build, quiet)
}

func (module *ModFiles) Requires(quiet bool) (descriptor.Requires, error) {
  return module.Descriptor.Requires(quiet)
}

func (module *ModFiles) BuildRequires(quiet bool) (descriptor.Requires, error) {
  return module.Descriptor.BuildRequires(quiet)
}

// Lists the source files of a Go package, without using the “go list” command
func (module *ModFiles) Sources(path string) error {

  p, serr := build.ImportDir(path, build.IgnoreVendor)
  if serr != nil {
    // Directories without any Go source file are not a problem ; other errors are
    if _, ok := serr.(*build.NoGoError) ; !ok {
      return serr
    }
  }
  for   _, fs := range [][]string {
                   p.GoFiles, p.CgoFiles,
                   // Without "IgnoredGoFiles" the result can not be used by
                   // downstreams that set specific non-arch build constrains
                   // such as “selinux”
                   p.IgnoredGoFiles,
                   p.CFiles, p.CXXFiles, p.MFiles, p.HFiles, p.FFiles, p.SFiles,
                   p.SwigFiles, p.SwigCXXFiles,
                   p.TestGoFiles } { // Tests are an integral part of Go modules
    for _, f  := range fs {
       rel, err := filepath.Rel(module.Root,filepath.Join(path,f))
       if err != nil {
         return err
       }
       module.Files[rel] = true
    }
  }
  return nil
}
