// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package read

import (
  "encoding/json"
  "errors"
  "io/ioutil"
  "os"
  "path/filepath"
  "regexp"
  "strings"
  "pagure.io/modist/version"
)

var ErrNotASymLink = errors.New("not a symlink")
var ErrOutsideSymlink  = errors.New("symlink pointing outside the module")

// Tells if the file located at path under root is a symbolic link,
// And if it points outside the root
func Symlink(root string, path string) (originalTarget string, modTarget string, simplifiedTarget string, err error) {
  fullPath  := filepath.Join(root, path)
  info, err := os.Lstat(fullPath)
  if err != nil {
    return  originalTarget, modTarget, simplifiedTarget, err
  }
  if info.Mode()&os.ModeSymlink == 0 {
   return  originalTarget, modTarget, simplifiedTarget, ErrNotASymLink
  }
  originalTarget, err = os.Readlink(fullPath)
  if err != nil {
    return  originalTarget, modTarget, simplifiedTarget, err
  }
  simplifiedTarget = originalTarget
  if !strings.HasPrefix(simplifiedTarget, "/") {
    simplifiedTarget = filepath.Join(filepath.Dir(fullPath), simplifiedTarget)
  }
  absTarget, err := filepath.Abs(simplifiedTarget)
  if err != nil {
    return originalTarget, modTarget, simplifiedTarget, err
  }
  absRoot, err := filepath.Abs(root)
  if err != nil {
    return originalTarget, modTarget, simplifiedTarget, err
  }
  modTarget, err = filepath.Rel(absRoot, absTarget)
  if err != nil || strings.HasPrefix(modTarget, "../") {
    return originalTarget, modTarget, simplifiedTarget, ErrOutsideSymlink
  }
  absPath, err := filepath.Abs(fullPath)
  if err != nil {
    return originalTarget, modTarget, simplifiedTarget, err
  }
  simplifiedTarget, err = filepath.Rel(filepath.Dir(absPath), absTarget)
  return originalTarget, modTarget, simplifiedTarget, err
}

var ErrBadExt   = errors.New("bad file extension")

// Try to infer a module version from the information present around the module
// descriptor path
func Version(path string, quiet bool) (*version.Version, error) {
  path = filepath.Clean(path)
  if filepath.Ext(path) != ".mod" {
    return nil, ErrBadExt
  }
  // Check if there is a matching info file lying around
  ip          := strings.TrimSuffix(path, ".mod")+ ".info"
  i, err      := Info(ip)
  if err == nil {
    result, err  := i.Ver(quiet)
    if err == nil {
      return result, err
    }
  }
  // Otherwise if we have a ~/go/pkg/mod/cache/download version.mod path
  r  := strings.TrimSuffix(filepath.Base(path), ".mod")
  result, err := version.FromGo(r, quiet)
  if err == nil {
    return result, err
  }
  // Otherwise try to infer from a ~/go/pkg/mod/ style qualified path first
  d, err      := filepath.Abs(filepath.Dir(path))
  vRe         := regexp.MustCompile(`.*@v(.*)$`)
  v           := vRe.ReplaceAllString(filepath.Base(d), "v${1}")
  result, err  = version.FromGo(v, quiet)
  if err == nil {
    return result, err
  }
  return result, err
}


// Parses a Go .info file
func Info(path string) (*version.Info, error) {
  path = filepath.Clean(path)
  if filepath.Ext(path) != ".info" {
    return nil, ErrBadExt
  }
  jsonFile, err := os.Open(path)
  if err != nil {
    return nil, err
  }
  byteValue, err := ioutil.ReadAll(jsonFile)
  if err != nil {
    return nil, err
  }
  var result version.Info
  err = json.Unmarshal(byteValue, &result)
  return &result, err
}
