// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package version

import (
  "errors"
  "fmt"
  "os"
  "regexp"
  "strconv"
  "strings"
  "time"
)

// Version syntax (including pseudo-versions) taken from
// src/cmd/go/internal/modload/help.go
// This whole part should be redone using
// https://godoc.org/cmd/go/internal/modfetch
// once this is exposed somewhere.
type Version struct {
  Epoch     int
  Version   string
  Pre       string
  Commit    string
  Time      time.Time
  Build     string
}

// Process raw with the exp regexp expression, containing named groups; return a
// map of the matched group values, indexed by the group names
func getSubs(raw string, exp string) (result map[string]string, err error) {
  result = make(map[string]string)
  re, err := regexp.Compile(exp)
  if err != nil {
    return nil, err
  }
  match := re.FindStringSubmatch(raw)
  if match != nil {
    for i, name := range re.SubexpNames() {
      if i != 0 && name != "" && match[i] != "" {
        result[name] = match[i]
      }
    }
  }
  return result, nil
}

func intString(raw string) (result string, err error) {
  i, err := strconv.Atoi(raw)
  result  = strconv.Itoa(i)
  return result, err

}

var ErrUnknownVersionFormat = errors.New("unknown version format")
var ErrBadVersionTime = errors.New("could not parse the version timestamp")

// Clean up preversion parts
func cleanPre(raw string) (result string) {
  result = "." + raw + "."

  forbiddenRe := regexp.MustCompile(`[^0-9A-Za-z.-]+`)
  result       = forbiddenRe.ReplaceAllString(result,"")

  leadingHyphen := regexp.MustCompile(`-[.-]+`)
  result         = leadingHyphen.ReplaceAllString(result, "-")

  leadingDot := regexp.MustCompile(`\.[.-]+`)
  result      = leadingDot.ReplaceAllString(result, ".")

  zeroLeadingNumberRe := regexp.MustCompile(`\.0+([1-9]\d*)\.`)
  result               = zeroLeadingNumberRe.ReplaceAllString(result, ".${1}.")

  result = strings.TrimPrefix(result, ".")
  result = strings.TrimSuffix(result, ".")
  result = strings.TrimSuffix(result, "-")
  result = strings.ToLower(result) // Pre-version end up in rpm filenames which should be case-insensitive
  return result
}

func New(version string, pre string, commit string, timestamp string,
         build string, epoch int, quiet bool) (result *Version, err error){
  var r Version
  result = &r
  if version != "" {
    vsubs, err := getSubs(version, `^v?(?P<major>\d+)(\.(?P<minor>\d+)(\.(?P<patch>\d+))?)?$`)
    if err != nil || len(vsubs) == 0 {
      return nil, ErrUnknownVersionFormat
    }
    major, _ := intString(vsubs["major"])
    minor := "0"
    _, exists := vsubs["minor"]
    if exists {
      minor, _ = intString(vsubs["minor"])
    }
    patch := "0"
    _, exists  = vsubs["patch"]
    if exists {
      patch, _ = intString(vsubs["patch"])
    }
    result.Version = major + "." + minor + "." + patch
  }
  result.Pre     = cleanPre(pre)
  if !quiet && result.Pre != pre {
    fmt.Fprintf(os.Stderr,"   WARNING converting “%s” pre-version suffix to “%s”.\n", pre, result.Pre)
  }
  result.Commit = cleanPre(commit)
  if !quiet && result.Commit != commit {
    fmt.Fprintf(os.Stderr,"   WARNING converting “%s” commit ID to “%s”.\n", commit, result.Commit)
  }
  if timestamp != "" {
    result.Time, err = time.Parse("20060102150405", timestamp)
    if err != nil {
      result.Time, err = time.Parse(time.RFC3339, timestamp)
      if err != nil {
        return result, ErrBadVersionTime
      }
    }
    result.Time = result.Time.UTC()
  }
  result.Build = build
  result.Epoch = epoch
  return result, nil
}

func FromGo(raw string, quiet bool) (result *Version, err error) {
  var r Version
  result = &r
  versionGr   := `(?P<version>\d+\.\d+\.\d+)`
  timeGr      := `(?P<time>\d{4}[0-1]\d[0-3]\d[0-2]\d[0-5]\d[0-5]\d)` // yyyymmddhhmmss
  preGr       := `(?P<pre>[[:alnum:]]([[:alnum:]-.]*[[:alnum:]])?)`
  commitGr    := `(?P<commit>[[:alnum:]-]{12})`
  buildGr     := `\+(?P<build>[[:alnum:]-.]+)`
  vsubs, err  := getSubs(raw, `^v` + versionGr + `((-((` + preGr + `\.)?0\.)?` + timeGr + `-` + commitGr + `)|(-` + preGr + `)?)(` + buildGr + `)?$`)
  if err != nil || len(vsubs) == 0 {
    return result, ErrUnknownVersionFormat
  }
  version, _   := vsubs["version"]
  pre, _       := vsubs["pre"]
  commit, _    := vsubs["commit"]
  timestamp, _ := vsubs["time"]
  return New(version, pre, commit, timestamp, "", 0, quiet)
}

// Target ordering and syntax, applying Go and RPM ordering rules
// Go version syntax (including pseudo-versions) taken from
// src/cmd/go/internal/modload/help.go and https://semver.org/
// RPM versionning rules taken from
// https://docs.fedoraproject.org/en-US/packaging-guidelines/Versioning/
// and can be cheched with rpmdev-vercmp
// Go                                       │ RPM
// vX.0.0-yyyymmddhhmmss-abcdefabcdef       │ X.0.0~0.yyyymmddhhmmss.abcdefabcdef
// vX.0.0-pre                               │ X.0.0~1.pre
// vX.0.0-pre.0.yyyymmddhhmmss-abcdefabcdef │ X.0.0~1.pre.yyyymmddhhmmss-abcdefabcdef
// vX.0.0                                   │ X.0.0
// vX.Y.Z-0.yyyymmddhhmmss-abcdefabcdef     │ X.Y.Z~0.yyyymmddhhmmss.abcdefabcdef
// vX.Y.Z-pre                               │ X.Y.Z~1.pre
// vX.Y.Z-pre.0.yyyymmddhhmmss-abcdefabcdef │ X.Y.Z~1.pre.yyyymmddhhmmss.abcdefabcdef
// vX.Y.Z                                   │ X.Y.Z
// vX.Y.(Z+1)-0.yyyymmddhhmmss-abcdefabcdef │ X.Y.(Z+1)~0.yyyymmddhhmmss.abcdefabcdef
// vX.Y.(Z+1)-pre                           │ X.Y.(Z+1)~1.pre
func (ver *Version) Go() (result string) {
  version := ver.Version
  if version == "" {
    version = "0"
  }
  for strings.Count(version, ".") < 2 {
    version = version + ".0"
  }
  time    := ver.Time.Format("20060102150405")
  result = "v" + version
  sep := "-"
  if ver.Pre != "" {
    result = result + sep + ver.Pre
    sep = "."
  }
  sup := ""
  if ver.Commit != "" {
    commit := string([]rune(ver.Commit)[:12])
    sup = time + "-" + commit
    if !strings.HasSuffix(version,".0.0") {
      sup = "0." + sup
    }
    result = result + sep + sup
  }
  return result
}

func (ver *Version) RPM() (result string) {
  version := ver.Version
  if version == "" {
    version = "0"
  }
  for strings.HasSuffix(version, ".0") {
    version = strings.TrimSuffix(version, ".0")
  }
  time    := ver.Time.Format("20060102150405")
  result = version
  sep := "~0."
  if ver.Pre != "" {
    result = result + "~1." + ver.Pre
    sep = "."
  }
  if ver.Commit != "" {
    commit := string([]rune(ver.Commit)[:12])
    result = result + sep + time + "." + commit
  }
  // rpm separates version from other naming components with hyphen
  // it is therefore forbidden within rpm version strings
  hyphenRe := regexp.MustCompile(`-`)
  result = hyphenRe.ReplaceAllString(result, ".")
  if ver.Epoch > 0 {
    result = strconv.Itoa(ver.Epoch) + ":" + result
  }
  if ver.Build != "" {
    result = result + "-" + ver.Build
  }
  return result
}

var ErrUnknownOutputFormat = errors.New("unknown output format")

func (ver *Version) Format(format string) (result string, err error) {
  switch format {
    case "rpm":
      return ver.RPM(), nil
    case "go":
      return ver.Go(),  nil
  }
  return "", ErrUnknownOutputFormat
}

// Taken from https://tip.golang.org/cmd/go/#hdr-Module_proxy_protocol that
// helpfuly states “this may be expanded in the future”
// Completed with semver & rpm parts Go developpers forgot to include
type Info struct {
  Version string                        // version string, in Go format
  Time    time.Time `json:",omitempty"` // code timestamp (unreliable)
  Build   string    `json:",omitempty"` // Build (semver) / Release (RPM)
  Epoch   int       `json:",omitempty"` // versioning Epoch (RPM)

}

func NewInfo(version string, Time time.Time, build string, epoch int) *Info {
  info := Info{Version: version,
               Time: Time,
               Build: build,
               Epoch: epoch}
  return &info
}

func (ver *Version) Info() *Info {
  return NewInfo(ver.Go(), ver.Time, ver.Build, ver.Epoch)
}

func (ver *Info) Ver(quiet bool) (*Version, error) {
  v, err := FromGo(ver.Version, quiet)
  v.Time  = ver.Time
  v.Build = ver.Build
  v.Epoch = ver.Epoch
  return v, err
}

