// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package write

import (
  "archive/zip"
  "bytes"
  "encoding/json"
  "errors"
  "fmt"
  "io"
  "os"
  "path/filepath"
  "sort"
  "strings"
  "syscall"
  "golang.org/x/mod/module"
  "golang.org/x/mod/sumdb/dirhash"
  "golang.org/x/sys/unix"
  "pagure.io/modist/descriptor"
  "pagure.io/modist/files"
  "pagure.io/modist/read"
  "pagure.io/modist/version"
)

// A map of produced file paths to a list special attributes such as
// "dir": the result is a directory
// "ghost": the result file content will need to be recomputed on installation
type Files map[string][]string

// Create a new file
// radical is the target path without extension
// reference is something that will be taken as modification time  reference
func (written Files) File(reader io.Reader, prefix string, radical string,
                          extension string, reference string, overwrite bool,
                         ) error {
  if extension != "" {
    extension   =  "." + strings.TrimPrefix(extension, ".")
  }
  pseudoPath := strings.TrimSuffix(radical, extension) + extension
  realPath   := filepath.Join(prefix, pseudoPath)
  if !overwrite {
    if _, err := os.Lstat(realPath); !os.IsNotExist(err) {
      return ErrWriteNotSafe
    }
  }
  file, err  := os.Create(realPath)
  if err != nil {
    return err
  }
  if filepath.Ext(pseudoPath) == ".md" {
    written[pseudoPath] = []string{"doc"}
  } else {
    written[pseudoPath] = nil
  }
  defer file.Close()
  if _, err = io.Copy(file, reader) ; err != nil {
    return err
  }
  if err := SetModTime(realPath, reference) ; err != nil {
    return err
  }
  return file.Sync()
}

// Create a new directory
// reference is something that will be taken as modification time  reference
// It is provided as a convenience, sicne writing anything in the directory
// will change its mtime; a saparate SetModTime call is usually necessary
func (written Files) Dir(prefix string, path string, reference string,
                         quiet bool) error {
  realPath   := filepath.Join(prefix, path)
  if !quiet {
    if _, err := os.Lstat(realPath); !os.IsNotExist(err) {
      fmt.Printf("     %s already exists!\n", realPath)
    }
  }
  if err := os.MkdirAll(realPath, 0777) ; err != nil {
    return err
  }
  written[path] = []string{"dir"}
  return SetModTime(filepath.Join(prefix, path), reference)
}

// Create a new symbolic link
// reference is something that will be taken as modification time  reference
func (written Files) Symlink(prefix string, path string, target string,
                           reference string) error {
  realPath   := filepath.Join(prefix, path)
  if _, err := os.Lstat(realPath); !os.IsNotExist(err) {
    return ErrWriteNotSafe
  }
  if err := os.Symlink(target,realPath) ; err != nil {
    return err
  }
  written[path] = nil
  return SetModTime(filepath.Join(prefix, path), reference)
}


// Create a new packed module description file
func (written Files) Descriptor(descriptor *descriptor.Descriptor,
                                prefix string, radical string,
                                reference string, quiet bool) error {
  if err := descriptor.Trim(quiet) ; err != nil {
    return err
  }
  data, err := descriptor.Format()
  if err != nil {
    return err
  }
  return written.File(bytes.NewReader(data), prefix, radical, "mod", reference,
                      false)
}

// Create a new module info file
func (written Files) Info(info *version.Version, prefix string,
                          radical string, reference string) error {
  data, err := json.MarshalIndent(info.Info(), "", "  ")
  if err != nil {
    return fmt.Errorf(
      "failed to marchal module revision information as json data,  %v", err)
  }
    return written.File(bytes.NewReader(data), prefix, radical, "info",
                        reference, false)
}

func zipFiles(writer io.Writer, module *files.Files, quiet bool) error {
  zipWriter := zip.NewWriter(writer)
  defer zipWriter.Close()
  paths := make(map[string]bool)
  for _, path := range module.Selected {
    paths[path] = true
  }
  dirs := make(map[string]bool)
  for _, path := range module.Selected {
    dirs[filepath.Dir(path)] = true
  }
  delete(dirs, ".")
  for dir := range dirs {
    for root := filepath.Dir(dir) ; root != "." ; root = filepath.Dir(root) {
      dirs[root] = true
    }
  }
  for _, path := range module.Selected {
    var file *os.File
    var mode  os.FileMode
    mode = 0644
    target  := ""
    spath := filepath.Join(module.Root, path)
    info, err := os.Lstat(spath)
    if err != nil {
      return err
    }
    target, mt, st, err := read.Symlink(module.Root, path)
    switch err {
      case read.ErrNotASymLink:
        file, err = os.Open(spath)
        if err != nil {
          return err
        }
      case read.ErrOutsideSymlink:
        fmt.Fprintf(os.Stderr,"   WARNING %s target %s is outside the module\n",
                    path, target)
        mode = 0755 | os.ModeSymlink
      case nil:
        if !quiet && target != st {
          fmt.Printf("     simplifying %s target %s to %s\n", path, target, st)
        }
        _, existsP := paths[mt]
        _, existsD := dirs[mt]
        if !existsP && !existsD {
          fmt.Fprintf(os.Stderr,
                      "   WARNING %s target %s is not included in the module\n",
                      path, st)
        }
        target = st
        mode = 0755 | os.ModeSymlink
      default:
        return err
    }
    header, err := zip.FileInfoHeader(info)
    if err != nil {
      return err
    }
    // Set a full path for the zipped file within the module
    header.Name = filepath.Join(module.Descriptor.Module.Mod.Path + "@" +
                                module.Version.Go(), path)
    // Set file compression
    // http://golang.org/pkg/archive/zip/#pkg-constants
    header.Method = zip.Deflate
    // Force uniform permissions in every packed module
    header.SetMode(mode)
    writer, err := zipWriter.CreateHeader(header)
    if err != nil {
      return err
    }
    if file != nil {
      _, err = io.Copy(writer, file)
    } else {
      _, err = io.Copy(writer, strings.NewReader(target))
    }
    if err != nil {
      return err
    }
  }
  zipWriter.Close()
  return nil
}

// Create a new packed module zip file
func (written Files) Zip(module *files.Files, prefix string,
                         radical string, quiet bool) error {
  reader, writer := io.Pipe()
  defer reader.Close()
  defer writer.Close()
  errors := make(chan error)
  go func() { errors <- written.File(reader, prefix, radical, "zip", "", false)
            } ()
  go func() { err := zipFiles(writer, module, quiet) ; writer.Close()
              errors <- err } ()
  for i := 0 ; i < 2 ; i++ {
    if err := <-errors ; err != nil {
      return err
    }
  }
  return nil
}

// Create a new packed module Hash file
func (written Files) Hash(prefix string, radical string) error {
  pseudoZipPath := strings.TrimSuffix(radical, ".ziphash") + ".zip"
  realZipPath   := filepath.Join(prefix, pseudoZipPath)
  data, err := dirhash.HashZip(realZipPath, dirhash.DefaultHash)
  if err != nil {
    return fmt.Errorf(
           "could not compute the hash of “%s” (packed module zip file)”, %v",
           realZipPath, err)
  }
  return written.File(strings.NewReader(data + "\n"), prefix, radical,
                      "ziphash", "", false)
}

func (written Files) Pack(mod *files.Files, prefix string,
                          destination string, quiet bool) error {
  path := mod.Descriptor.Module.Mod.Path
  encPath, err := module.EscapePath(path)
  if err != nil {
    return err
  }
  encVer, err := module.EscapeVersion(mod.Version.Go())
  if err != nil {
    return err
  }
  relDestination  := filepath.Join(encPath , "@v")
  fullDestination := filepath.Join(destination, relDestination)
  err = os.MkdirAll(filepath.Join(prefix,fullDestination), 0777)
  if err != nil {
    return err
  }
  for relDestination != "." {
    written[filepath.Join(destination, relDestination)] = []string{"dir"}
    relDestination = filepath.Dir(relDestination)
  }
  radical := filepath.Join(fullDestination, encVer)
  if err := written.Zip(mod, prefix, radical, quiet) ; err != nil {
    return err
  }
  if err := written.Hash(prefix, radical) ; err != nil {
    return err
  }
  err = written.Descriptor(mod.Descriptor, prefix, radical, "", quiet)
  if err != nil {
    return err
  }
  err = written.Info(mod.Version, prefix, radical, "")
  written[filepath.Join(fullDestination, "list")] = []string{"ghost"}
  return err
}

var ErrWriteNotSafe = errors.New("destination may already be in use")

// Sets the modification time of path to the same as reference
func SetModTime(path string, reference string) (err error) {
  if reference != "" {
    info, err := os.Lstat(reference)
    if os.IsNotExist(err) {
      return nil
    } else if err != nil {
      return err
    }
    // The following does not seem to work on directories with subdirectories
    stat := info.Sys().(*syscall.Stat_t)
    atime := unix.NsecToTimeval(syscall.TimespecToNsec(stat.Atim))
    mtime := unix.NsecToTimeval(syscall.TimespecToNsec(stat.Mtim))
    err = unix.Lutimes(path, []unix.Timeval{atime, mtime})
    return err
  }
  return nil
}

func (written Files) Text(prefix string, special bool) string {
  var paths []string
  for path := range written {
    paths = append(paths, path)
  }
  sort.Strings(paths)
  var data string
  for _, path := range paths {
    if written[path] == nil || special {
      data = data + filepath.Join(prefix, path) + "\n"
    }
  }
  return data
}

func (written Files) RPM() string {
  var paths []string
  for path := range written {
    paths = append(paths, path)
  }
  sort.Strings(paths)
  var data string
  for _, path := range paths {
    for _, mod := range written[path] {
      data = data + "%" + mod + " "
    }
    data = data + path + "\n"
  }
  return data
}

func (written Files) JSON() (string, error) {
  data, err := json.MarshalIndent(written, "", "  ")
  if err != nil {
     return "", fmt.Errorf("failed to marchal results as json data,  %v", err)
  }
  return string(data[:]), nil
}

var ErrUnknownOutputFormat = errors.New("unknown output format")

func (written Files) Format(format string) (string, error) {
  switch format {
    case "json":
      return written.JSON()
    case "rpm":
      return written.RPM(), nil
    case "txt":
      return written.Text("", true), nil
    default:
      return "", ErrUnknownOutputFormat
  }
}

func (written Files) Output(file string, format string) error {
  data, err := written.Format(format)
  if err != nil {
    return err
  }
  ext := format
  if format == "rpm" {
    ext = "lst"
  }
  return written.File(strings.NewReader(data), "", file, ext, "", true)
}
